#!/usr/bin/env python
#  -*- coding: utf-8 -*-
import ctpapi
import time as tm
import zhuchannel
import threading,queue
import os
import asyncio
import requests
import smtplib                                 # smtp服务器
from email.mime.text import MIMEText#发送文本
from email.mime.multipart import MIMEMultipart#生成多个部分的邮件体
from email.mime.application import MIMEApplication#发送图片
from concurrent.futures import ThreadPoolExecutor
import traceback,json
import types
from functools import partial
import polars
from datetime import datetime,time,date,timedelta
import copy
from typing import Dict, List, Optional, Tuple, Any, Union, Callable, Set
from zhustruct import Quote,Account,Position,Trade,Order,InstrumentProperty,CTPDataProcessor,MutableKlineHolder,MutableTickHolder

class PeopleQuantApi():
    def __init__(self,BrokerID:str,UserID:str,PassWord:str,AppID:str="", AuthCode:str="",TradeFrontAddr:str="",MdFrontAddr:str="",
                 s:str="",vip_code="",flowfile="" ,quote_queue:zhuchannel.ThreadChan=None,save_tick=False,ProductionMode=True,
                 _FTDmaxsize={"Session":6,"ReqQry":1,"OrderInsert":6,"OrderAction":6,"order_exe":10,"ex_order_exe":10},**kw):
        '''
        BrokerID:期货公司代码
        UserID:期货账号
        PassWord:账号密码
        AppID:穿透式认证程序编号
        AuthCode:穿透式认证授权码
        TradeFrontAddr:交易前置地址
        MdFrontAddr:行情前置地址
        s:日志文件夹名称
        flowfile:数据流文件保存目录,默认在当前程序目录下创建
        quote_queue:行情队列,用于从CTP行情接口接收行情,推送向策略层(策略层若需要CTP行情)
        save_tick:是否保存tick到本地
        ProductionMode:True:生产模式,False:测试模式
        '''
        self._subscribequeue = zhuchannel.ThreadChan() #mdapi从各层包括tradeapi接收订阅、退订
        self._quote_queue = zhuchannel.ThreadChan() #mdapi向tradeapi发送行情
        self.quote_queue = quote_queue #各层从mdapi接收行情
        self._Reqqueue = zhuchannel.ThreadChan() #请求队列
        self.MarketDataqueue = []
        if self.quote_queue is not None:
            self.MarketDataqueue.append(quote_queue)
        self.instruments = set()  #全部合约代码
        self.quote = Quote()  #全部合约行情
        self.instruments_property = {} #合约属性
        self.positions = Position()  #全部合约持仓
        self.orders = Order()  #全部委托单
        self.trades = Trade()  #全部成交单
        self.account= Account()  #账户资金
        self.subscribe_instruments = set() #已订阅合约代码
        self.unsubscribe_instruments = set() #已退订合约代码
        self._exception_queue = zhuchannel.ThreadChan()  # 用于传递异常的队列
        self._rtn_queue = zhuchannel.ThreadChan()  # 用于接收CTP主动发出的回报队列
        self._ans_queue = zhuchannel.ThreadChan( )  # 用于接收CTP反馈的结果队列
        self._ReqID = 0  #请求编号
        self._save_tick = save_tick
        self._save_tick_symbols = {}
        self._save_tick_queue = zhuchannel.ThreadChan( )  # 用于保存ticks队列
        self._pos_queue :Dict[str, Union[asyncio.Queue,queue.Queue]] = {}
        self._FTDmaxsize = _FTDmaxsize
        if not flowfile:
            try:
                self._flowfile = os.path.dirname(os.path.abspath(__file__)) #当前程序目录(该py文件__file__所在目录)
            except: self._flowfile = os.getcwd()   #程序工作目录下创建目录 
        else: self._flowfile = flowfile
        self._logfile = fr"{self._flowfile}\logs\{BrokerID}\{UserID}\{s}"
        self.CTP_data_processor = CTPDataProcessor(_logfile=self._logfile)

        self._tradekwargs = {"BrokerID":BrokerID,"UserID":UserID,"Password":PassWord,"AppID":AppID, "AuthCode":AuthCode,"s":s,"flowfile":self._flowfile,"Subscribequeue":self._subscribequeue,
                             "quote_queue":self._quote_queue,"Reqqueue":self._Reqqueue,"Notify":{"rtn_queue":self._rtn_queue,"ans_queue":self._ans_queue},"ProductionMode":ProductionMode,
                             "TradeFrontAddr":TradeFrontAddr,"PQexception_queue":self._exception_queue,"MarketDataqueue":self.MarketDataqueue,"_FTDmaxsize":self._FTDmaxsize,"vip_code":vip_code,**kw}
        self._mdkwargs = {"BrokerID":BrokerID,"UserID":UserID,"Password":PassWord,"AppID":AppID, "AuthCode":AuthCode,"s":s,"flowfile":self._flowfile,"Subscribequeue":self._subscribequeue,
                          "quote_queue":self._quote_queue,"Notify":{"rtn_queue":self._rtn_queue,"ans_queue":self._ans_queue},"ProductionMode":ProductionMode,"_FTDmaxsize":self._FTDmaxsize,
                          "MdFrontAddr":MdFrontAddr,"vip_code":vip_code,**kw}
        
        self.tradethread = zhuchannel.WorkThread(target=ctpapi.TraderApi,args=tuple(),kwargs=self._tradekwargs,_api=True,name_prefix="tradethread")
        self.mdthread = zhuchannel.WorkThread(target=ctpapi.MdApi,args=tuple(),kwargs=self._mdkwargs,_api=True,name_prefix="mdthread")
        self.tradethread.start()
        self.mdthread.start()
        self.rtnthread = zhuchannel.WorkThread(target=self._rtn_thread,args=tuple(),kwargs={},exception_queue=self._exception_queue )
        self.rtnthread.start()
        self.update_tick_kline = zhuchannel.WorkThread(target=self._update_tick_kline,args=tuple(),kwargs={},exception_queue=self._exception_queue )
        self.update_tick_kline.start()
        self.exception = zhuchannel.WorkThread(target=self.exception_thread,args=tuple(),kwargs={} )
        self.exception.start()
        self.executor = ThreadPoolExecutor()
        try:
            r1 = self._ans_queue.get(timeout=120) #等待交易初始化完成
            r2 = self._ans_queue.get(timeout=120) #等待行情初始化完成
            #self.TradingDay = self._get_trading_day()
            print("交易接口,行情接口:初始化完成")
        except Exception :
            e = "初始化失败,可能网络连接超时,请检查网络"
            self._exception_queue.put(e)
            raise Exception(e)
    def Join(self):
        self.tradethread._Join()
        self.mdthread._Join()
    def exception_thread(self):
        for e in self._exception_queue:
            e = f'\n{datetime.today()}{"-"*30}+\n{e}\n'
            self.logs_txt(e,self.tradethread.api._logfile)
            if self.rtnthread.is_alive(): self._rtn_queue.put("exception")
            self.Join()
            self.tradethread.join()
            self.mdthread.join()
            if 0:
                print(self.tradethread.is_alive(),self.mdthread.is_alive())
                threads = threading.enumerate()
                for idx, thread in enumerate(threads):
                    print(f"线程 {idx + 1}:")
                    print(f"  名称: {thread.name}")
                    print(f"  ID: {thread.ident}")  # 线程ID（可能为None，未启动时）
                    print(f"  是否存活: {thread.is_alive()}")
                    print(f"  是否为守护线程: {thread.daemon}")
                #raise Exception(e)
                for t in self.threads:
                        t.join(timeout=1.0)
                        if t.is_alive():
                            print(f"线程 {t.name} 无法优雅终止，已强制中断")
                        else:
                            print(t.is_alive(),t.name,"死了")
    
    def _get_trading_day(self,wait_return=False) -> str:
        '''查询当前交易日'''
        r = self._sendReq({"reqfuncname":"_get_trading_day","wait_return":wait_return, })
        if not r['ret'] : print("查询交易日失败")
        else: return r['ret']

    def get_account(self,wait_return=False) -> Account:
        '''
        获取账户信息
        Return:
            账户字典
        '''
        if self.account: return self.account
        else:
            r = self._sendReq({"reqfuncname":"get_account","wait_return":wait_return, })
            if not r['ret'] : 
                e = f"{datetime.now()} -get_account查询账户失败\n"
                self.logs_txt(e,self._logfile,_print=True)
            else: 
                self.account.update( r['ret'] )
                return self.account
            
    def get_position(self,InstrumentID:str="",wait_return=False,_ctp = None) -> Union[Dict[str,Position] ,Position]:
        '''
        获取品种持仓,本地计算
        Args:
            InstrumentID: 合约码
        Return:
            若查询不到返回缺省值
            返回合约的持仓字典,包含健:
                "pos_long", "pos_long_today", "pos_long_his","open_price_long" ,"position_price_long",
                "pos_short", "pos_short_today", "pos_short_his","open_price_short","position_price_short",
                "position_profit_long","float_profit_long","position_profit_short",
                "float_profit_short","margin","exch_margin",
                "margin_long","margin_short","margin_rate_long","margin_rate_short",
                "margin_volume_long","margin_volume_short",
                "instrument_id","exchange_id","ins_class","strike_price",
                "short_frozen_today","long_frozen_today","short_frozen_his","long_frozen_his",
                "OpenRatioByMoney","OpenRatioByVolume","CloseRatioByMoney","CloseRatioByVolume",
                "CloseTodayRatioByMoney","CloseTodayRatioByVolume",
            不填合约码则返回全部账户嵌套持仓字典,字典健为合约码,值为持仓字典
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        if InstrumentID in self.positions:
            return self.positions[InstrumentID]
        else:
            self.get_quote(InstrumentID)
            r = self._sendReq({"reqfuncname":"get_position","InstrumentID":InstrumentID,"wait_return":wait_return,"_ctp":_ctp, })
            if InstrumentID:
                self.positions[InstrumentID] = Position( )
                self.positions[InstrumentID].update(r['ret'])
                return self.positions[InstrumentID]
            else: 
                for s in r['ret']:
                    if s not in self.positions: self.positions[s] = Position( )
                    self.positions[s].update(r['ret'][s])
                return self.positions
    
    def get_quote(self,InstrumentID:str,wait_return=True,subscribe=True) -> Quote:
        '''
        Args:
            InstrumentID: 合约码
        Return:
            返回合约的行情Quote对象
        '''
        if not self.check_instrument(InstrumentID): return
        if InstrumentID not in self.instruments_property:
            self.get_symbol_info(InstrumentID)
        if InstrumentID in self.quote and InstrumentID not in self.unsubscribe_instruments:
            return self.quote[InstrumentID]
        elif InstrumentID:
            if subscribe: 
                q = self.subscribe_quote([InstrumentID]) #订阅行情
            r = self._sendReq({"reqfuncname":"get_quote","InstrumentID":InstrumentID,"wait_return":wait_return, })
            if not r['ret'] : 
                e = f"{datetime.now()} -get_quote查询行情失败\n"
                self.logs_txt(e,self._logfile,_print=True)
            else:
                while InstrumentID not in self.quote:  #等待行情更新
                    tm.sleep(0.001)
                else: return self.quote[InstrumentID]

    def get_symbol_trade(self,InstrumentID:str = "",wait_return=False) -> List[Trade]:
        '''
        获取合约成交单
        Ags:
            InstrumentID: 合约代码,如rb2601,不填则查询全部成交单
        Return:
            由合约的成交单Trade对象组成的列表
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_symbol_trade","InstrumentID":InstrumentID,"wait_return":wait_return})
        if not r['ret']: 
            e = f"{datetime.now()} -get_symbol_trade查询成交单为空\n"
            self.logs_txt(e,self._logfile,_print=True)
            return []
        return [Trade().update(t).update({"local_timestamp":tm.time()}) for t in r['ret']]

    def get_symbol_order(self,InstrumentID:str = "", OrderStatus:str = "",CombOffsetFlag:str="",wait_return=False) -> List[Order]:
        '''
        获取合约委托单
        Ags:
            InstrumentID: 合约代码,如rb2601,不填则查询全部委托单
            OrderStatus: 委托单状态,"Alive"查询活动委托单,"Finished"查询已完结委托单
            CombOffsetFlag: Open开仓单,Close平仓单
        Return:
            由委托单Order对象组成的列表
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_symbol_order","InstrumentID":InstrumentID,"OrderStatus":OrderStatus,"CombOffsetFlag":CombOffsetFlag,"wait_return":wait_return })
        if not r['ret']: 
            e = f"{datetime.now()} -get_symbol_order查询委托单为空\n"
            self.logs_txt(e,self._logfile,_print=True)
            return []
        return [Order().update(o).update({"local_timestamp":tm.time()}) for o in r['ret']]
            
    def get_id_order(self,InstrumentID:str="", order_id:str="",OrderMemo:str="",wait_return=False) -> Union[List[Order] , Order]:
        '''
        获取合约委托单
        Ags:
            InstrumentID: 合约代码,如rb2601
            order_id: 为委托单order编号组成的字符串:f'{order["FrontID"]}_{order["SessionID"]}_{order["OrderRef"]}'
        Return:
            order_id有效时返回委托单对象,order_id不设置时返回委托单对象组成的列表
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        if order_id in self.orders:
            return self.orders[order_id]
        else:
            r = self._sendReq({"reqfuncname":"get_id_order","InstrumentID":InstrumentID,"order_id":order_id,"OrderMemo":OrderMemo,"wait_return":wait_return, })
            if isinstance(r['ret'],dict): 
                self.orders[order_id] = Order( )
                self.orders[order_id].update(r['ret'])
                self.orders[order_id].update({"local_timestamp":tm.time()})
                return self.orders[order_id]
            elif isinstance(r['ret'],list):
                if not r['ret']: 
                    e = f"{datetime.now()} -get_id_order查询委托单为空\n"
                    self.logs_txt(e,self._logfile,_print=True)
                else: 
                    for o in r['ret']:
                        if o["order_id"] not in self.orders: 
                            self.orders[o["order_id"]] = Order( )
                            self.orders[o["order_id"]].update(o)
                            self.orders[o["order_id"]].update({"local_timestamp":tm.time()})
                    if order_id: return self.orders[order_id]
                    return [self.orders[o["order_id"]] for o in r['ret']]
                       
    def get_trade_of_order(self, order_id:str,wait_return=False,_print=True) -> List[Trade]:
        '''
        获取合约委托单对应的成交单
        Ags:
            order_id: 必填,委托单order_id
        Return:
            由成交单字典组成的列表
        '''
        r = self._sendReq({"reqfuncname":"get_trade_of_order","order_id":order_id,"wait_return":wait_return, })
        if not r['ret']: 
            if _print:
                e = f"{datetime.now()} -get_trade_of_order查询成交单失败\n"
                self.logs_txt(e,self._logfile,_print=_print)
            return []
        else:
            return [Trade().update(t).update({"local_timestamp":tm.time()}) for t in r['ret']]
    
    def get_order_risk(self,InstrumentID:str = "",wait_return=False) -> dict:
        '''
        获取合约委托单风控统计
        Ags:
            InstrumentID: 合约代码,如rb2601
        Return:
            字典,健为:
                "cancel_count","open_volume","order_count",
                "self_trade_count","order_exe","otr","repeat"
                报单次数、撤单次数、多空开仓成交总手数、自成交数、信息量、报单成交比
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_order_risk","InstrumentID":InstrumentID,"wait_return":wait_return, })
        if not r['ret']: 
            e = f"{datetime.now()} -get_order_risk查询委托单风控失败\n"
            self.logs_txt(e,self._logfile,_print=True)
        return r['ret']

    def get_frozen_pos(self,InstrumentID:str,wait_return=False) -> dict:
        '''
        获取合约冻结仓位,上期所、能源中心区分昨仓、今仓,其他交易所只有昨仓冻结(体现在今仓冻结字段)
        由活动委托单计算得出
        Return:
            {"short_frozen_today":0,"long_frozen_today":0,
            "short_frozen_his":0,"long_frozen_his":0}
        '''
        if not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_frozen_pos","InstrumentID":InstrumentID,"wait_return":wait_return, })
        if not r['ret']: 
            e = f"{datetime.now()} -get_frozen_pos查询冻结持仓失败\n"
            self.logs_txt(e,self._logfile,_print=True)
        return r['ret']
    
    def get_frozen_margin(self,InstrumentID:str="",wait_return=False) -> float:
        '''
        获取冻结保证金,按多空活动开仓单保证金求和,默认多空开仓单都占用保证金,不考虑保证金优惠
        Args:
            InstrumentID: 合约代码,如rb2601,不填则查询账户冻结保证金
        Return:
            保证金之和
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_frozen_margin","InstrumentID":InstrumentID,"wait_return":wait_return, })
        if not r['ret']: 
            e = f"{datetime.now()} -get_frozen_margin查询冻结保证金失败\n"
            self.logs_txt(e,self._logfile,_print=True)
        return r['ret']
        
    def get_symbol_info(self,InstrumentID:str,wait_return=False) -> InstrumentProperty:
        '''
        获取单个具体合约(期货或期权)属性
        Args:
            InstrumentID:str,合约代码,若查询合约不存在返回空字典
        Return:
            合约属性对象InstrumentProperty
            若查询不存在返回None
        '''
        if not self.check_instrument(InstrumentID): return
        if InstrumentID in self.instruments_property:
            return self.instruments_property[InstrumentID]
        else:
            r = self._sendReq({"reqfuncname":"get_symbol_info","InstrumentID":InstrumentID,"wait_return":wait_return, })
            self.instruments_property[InstrumentID] = InstrumentProperty()
            self.instruments_property[InstrumentID].update(r['ret'])
            if self.instruments_property[InstrumentID]['OptionsType'] == '1': self.instruments_property[InstrumentID]['OptionsType'] = 'CALL'
            elif self.instruments_property[InstrumentID]['OptionsType'] == '2': self.instruments_property[InstrumentID]['OptionsType'] = 'PUT'
            return self.instruments_property[InstrumentID]
    def query_symbol(self,ProductID:str="",ExchangeID:str="",ProductClass:str="",OptionsType:str="") -> polars.DataFrame:
        '''
        根据交易所ID或品种ID查询合约属性,可查询期货、期权合约属性
        Args:
            ProductID:str,品种ID,如rb
            ExchangeID:str,交易所代码,上期所SHFE,大商所DCE,郑商所CZCE,能源中心INE,中金所CFFEX,广期所GFEX
                    不填则查找全部交易所
            ProductClass:str,查找合约类型,期货"Future",期权"Option"
            OptionsType:str,期权类型,看涨期权CALL,看跌期权PUT
        Return:
            polars.DataFrame 若查询不存在返回空表格
        '''
        r = self._sendReq({"reqfuncname":"query_symbol","ExchangeID":ExchangeID,"ProductID":ProductID,"ProductClass":ProductClass,
                            "OptionsType":OptionsType,})
        return r['ret']
    def query_symbol_option(self,UnderlyingInstrID:str,OptionsType:str) -> List[polars.DataFrame]:
        '''
        根据标的合约查询上市中的期权合约属性
        Args:
            UnderlyingInstrID:str,标的合约码
            OptionsType:str,期权类型,"CALL"看涨期权,"PUT"看跌期权
        Return:
            [ polars.DataFrame表格 , polars.DataFrame表格 ]
            返回不同到期日期权组成的列表,元素为polars.DataFrame表格
            若查询不存在返回空列表
            表格形式为:
            ┌───────────┬─────────────────┬────────────┬───────────────┬───┬─────────────┬───────────────────┬────────────────────┬────────────────┐
            │ row_index ┆ CombinationType ┆ CreateDate ┆ DeliveryMonth ┆ … ┆ StrikePrice ┆ UnderlyingInstrID ┆ UnderlyingMultiple ┆ VolumeMultiple │
            │ ---       ┆ ---             ┆ ---        ┆ ---           ┆   ┆ ---         ┆ ---               ┆ ---                ┆ ---            │
            │ u32       ┆ str             ┆ i64        ┆ i64           ┆   ┆ f64         ┆ str               ┆ f64                ┆ i64            │
            ╞═══════════╪═════════════════╪════════════╪═══════════════╪═══╪═════════════╪═══════════════════╪════════════════════╪════════════════╡
            │ 0         ┆ 0               ┆ 20250612   ┆ 10            ┆ … ┆ 5000.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 1         ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 5100.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 2         ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 5200.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 3         ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 5300.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 4         ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 5400.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ …         ┆ …               ┆ …          ┆ …             ┆ … ┆ …           ┆ …                 ┆ …                  ┆ …              │
            │ 9         ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 5900.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 10        ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 6000.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 11        ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 6100.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 12        ┆ 0               ┆ 20250605   ┆ 10            ┆ … ┆ 6200.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            │ 13        ┆ 0               ┆ 20250714   ┆ 10            ┆ … ┆ 6300.0      ┆ SR511             ┆ 1.0                ┆ 10             │
            └───────────┴─────────────────┴────────────┴───────────────┴───┴─────────────┴───────────────────┴────────────────────┴────────────────┘
        '''
        r = self._sendReq({"reqfuncname":"query_symbol_option","UnderlyingInstrID":UnderlyingInstrID,"OptionsType":OptionsType,})
        if not r['ret']: 
            e = f"{datetime.now()} -query_symbol_option查询期权为空:{r['ret']}\n"
            self.logs_txt(e,self._logfile,_print=True)
        return r['ret']
    def get_option(self,underlying_price,price_level,group_option:polars.DataFrame) -> dict:
        '''
        查询以价格underlying_price为基准的档位price_level对应的期权
        Args:
            underlying_price:float,标的价格
            price_level:int,期权档位,正值实值期权,负值虚值期权,0平值期权
            group_option: 由query_symbol_option返回的期权表格
        Return:
            {"option":SR511P5500,"strike_price":5500.0,"spread":100.0}
        '''
        r = self._sendReq({"reqfuncname":"get_option","underlying_price":underlying_price,
                            "price_level":price_level,"group_option":group_option,})
        return r['ret']

    def get_symbol_marginrate(self,InstrumentID:str,wait_return=False) -> dict:
        '''
        获取合约保证金率(非期权)
        Args:
            InstrumentID: 合约代码,如rb2601
        Return:
            若查询不到返回缺省值0
            保证金率字典,包含健 "LongMarginRatioByMoney", "ShortMarginRatioByMoney",
            "LongMarginRatioByVolume", "ShortMarginRatioByVolume",
            "LongMarginRatio","ShortMarginRatio"
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_symbol_marginrate","InstrumentID":InstrumentID,"wait_return":wait_return, })
        return r['ret']

    def get_symbol_commission(self,InstrumentID:str,wait_return=False) -> dict:
        '''
        获取合约手续费率(非期权)
        Args:
            InstrumentID: 合约代码,如rb2601
        Return:
            若查询不到返回缺省值0
            手续费率字典,包含健 "OpenRatioByMoney","OpenRatioByVolume","CloseRatioByMoney",
            "CloseRatioByVolume","CloseTodayRatioByMoney","CloseTodayRatioByVolume",
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_symbol_commission","InstrumentID":InstrumentID,"wait_return":wait_return, })
        return r['ret']
    
    def get_trade_commission(self,InstrumentID:str,TradeID:str="",wait_return=False) -> float:
        '''
        Args:
            InstrumentID: 合约代码,如rb2601
            TradeID: 成交单编号,若不填则计算合约全部成交单手续费
        Return:
            float,手续费总和
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        if TradeID and TradeID not in self.trades: return 
        r = self._sendReq({"reqfuncname":"get_trade_commission","InstrumentID":InstrumentID,"TradeID":TradeID,"wait_return":wait_return, })
        #if not r['ret']: print("查询成交单手续费失败")
        return r['ret']

    def get_option_commission(self,InstrumentID:str="", wait_return=False):
        '''
        获取期权合约手续费率
        Args:
            InstrumentID: 合约代码,如rb2601-P-3500
        Return:
            若查询不到返回缺省值0
            手续费率字典,包含健 
            "OpenRatioByMoney","OpenRatioByVolume","CloseRatioByMoney","CloseRatioByVolume",
            "CloseTodayRatioByMoney","CloseTodayRatioByVolume","StrikeRatioByMoney","StrikeRatioByVolume":0,
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_option_commission","InstrumentID":InstrumentID, "wait_return":wait_return, })
        return r['ret']

    def get_option_marginrate(self,InstrumentID:str,InputPrice,UnderlyingPrice=0,HedgeFlag='1',wait_return=False) -> dict:
        '''
        获取期权合约保证金率 
        #保证金=max(权利金+FixedMargin,MiniMargin)
        #每手 期权卖方交易保证金 = 权利金 + max(标的期货合约保证金 - 期权虚值额的一半，标的期货合约保证金的一半)
        #max(标的期货合约保证金 - 期权虚值额的一半 , 标的期货合约保证金的一半)为不变量,均以结算价和行权价计算
        Args:
            InstrumentID: 合约代码,如rb2601P3500
            InputPrice: 期权合约报价
            UnderlyingPrice： 标的合约价格,默认昨结算价
        Return:
            若查询不到返回缺省值0
            保证金率字典,包含健 
            "FixedMargin","MiniMargin","Royalty","ExchFixedMargin","ExchMiniMargin","LongMarginRatioByMoney",
            "ShortMarginRatioByMoney","LongMarginRatioByVolume","ShortMarginRatioByVolume","LongMarginRatio",
            "ShortMarginRatio",
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_option_marginrate","InstrumentID":InstrumentID,"InputPrice":InputPrice,"UnderlyingPrice":UnderlyingPrice,"HedgeFlag":HedgeFlag,"wait_return":wait_return, })
        return r['ret']
    
    def get_option_magin(self,InstrumentID,InputPrice,UnderlyingPrice=0,HedgeFlag='1',wait_return=False):
        '''
        计算期权保证金,变化的只有权利金,由InputPrice计算,当未订阅行情时,默认以行权价计算
        只是估计计算，因为其使用的公式()保证金=max(权利金+FixedMargin,MiniMargin))中的权利金部分在计算时使用的期权价格是InputPrice。
        而资金查询里的期权保证金计算公式中的期权价格是使用max算法(max(昨结算，最新价))得到的
        Args:
            InstrumentID: 合约代码,如rb2601P3500
            InputPrice: 期权合约报价
            UnderlyingPrice： 标的合约价格,默认昨结算价
        Return:
            float,保证金
        '''
        if InstrumentID and not self.check_instrument(InstrumentID): return
        r = self._sendReq({"reqfuncname":"get_option_magin","InstrumentID":InstrumentID,"InputPrice":InputPrice,"UnderlyingPrice":UnderlyingPrice,"HedgeFlag":HedgeFlag,"wait_return":wait_return, })
        return r['ret']

    def _wait_order(self,ReqID:int):
        try:
            r = self._ans_queue.get_nowait()
            if ReqID in r: return r[ReqID] #查询完成
            else: self._ans_queue.put_nowait(r)
        except queue.Empty:
            return True
 
    def cancel_order(self, order_id:str,OrderMemo="",wait_return=False):
        '''撤单阻塞函数,等待撤单结束'''
        if order_id not in self.orders:
            print("委托单号不存在")
            return
        if order_id in self.orders and self.orders[order_id]["OrderStatus"] in ['0','2','4' ,'5' ]: return
        r = self._sendReq({"reqfuncname":"cancel_order","order_id":order_id,"OrderMemo":OrderMemo,"wait_return":wait_return,})
        if not r['ret'] or isinstance(r['ret'],str) :
            e = f"{datetime.now()} -cancel_order撤单失败,order_id:{order_id},错误信息:{r['ret']}\n"
            self.logs_txt(e,self._logfile,_print=True)
            return
        else: #撤单发出但不代表被交易所接受
            e = f"{datetime.now()} -cancel_order已撤单,order_id:{order_id}\n"
            self.logs_txt(e,self._logfile,_print=True)
            return

    def cancel_all_order(self,InstrumentID:str="",OrderMemo: str = "",wait_return=False):
        '''批量撤单,阻塞函数,等待撤单结束'''
        orders = []
        if InstrumentID :
            if not self.check_instrument(InstrumentID): return
            orders = self.get_symbol_order(InstrumentID,"Alive")
        else: orders = self.get_symbol_order(OrderStatus = "Alive")
        if orders:
            for o in orders:
                self.cancel_order(o["order_id"],OrderMemo=OrderMemo)
        
    def insert_order(self,ExchangeID:str,InstrumentID:str,Direction:str,Offset:str,Volume:int,LimitPrice:float,advanced=None,HedgeFlag:str="1",WaitReturn=False,OrderMemo="") -> Union[None,Order]:
        '''非阻塞函数,报单被交易所接受后返回字典'''
        if not self.check_instrument(InstrumentID): return
        quote = self.get_quote(InstrumentID)
        instrument_property = self.get_symbol_info(InstrumentID)
        exchange_id = instrument_property["ExchangeID"]
        PriceTick = instrument_property["PriceTick"]
        #VolumeMultiple = instrument_property["VolumeMultiple"]
        if (not isinstance(LimitPrice,float) or LimitPrice != LimitPrice or not isinstance(Volume,int) or LimitPrice/PriceTick - int(LimitPrice/PriceTick) != 0 or
            Volume <= 0 or Volume < instrument_property["MinLimitOrderVolume"] or Volume > instrument_property["MaxLimitOrderVolume"] 
            or Direction not in ["Buy","Sell"] or Offset not in ["Open","Close","CloseToday"] or ExchangeID != exchange_id
            or (LimitPrice > quote["UpperLimitPrice"] and Direction == "Buy" or LimitPrice < quote["LowerLimitPrice"] and Direction == "Sell")): 
            e = ("{},{},{}\n".format(
                f"{datetime.now()} -insert_order报单失败,输入参数不合法,价格:{LimitPrice},数量:{Volume},最小限价单数量:{instrument_property['MinLimitOrderVolume']}", 
                f"最大限价单数量:{instrument_property['MaxLimitOrderVolume']},跌停价:{quote['LowerLimitPrice']},涨停价:{quote['UpperLimitPrice']}",
                f"最小跳:{PriceTick},交易方向:{Direction},开平:{Offset},合约交易所:{exchange_id}"
            ))
            self.logs_txt(e,self._logfile,_print=True)
            return e
        r = self._sendReq({"reqfuncname":"insert_order","ExchangeID":ExchangeID,"InstrumentID":InstrumentID,"Direction":Direction,"Offset":Offset,"Volume":Volume,"LimitPrice":LimitPrice,
                            "advanced":advanced,"HedgeFlag":HedgeFlag,"WaitReturn":WaitReturn,"OrderMemo":OrderMemo})
        if not isinstance(r['ret'],dict) :
            e = f"{datetime.now()} -insert_order报单失败,客户端拒绝或报单错误,错误信息:{r['ret']}\n"
            self.logs_txt(e,self._logfile,_print=True)
            return r['ret']
        else:  #{order_id:order}
            for order_id in r['ret']:
                if order_id not in self.orders: 
                    self.orders[order_id] = Order( )
                self.orders[order_id].update(r['ret'][order_id])
                self.orders[order_id].update({"local_timestamp":tm.time()})
            #报单发出成功
            return self.orders[order_id]

    def open_close(self,symbol,kaiping='',lot=0,price=None,block=True,n_price_tick=1,che_time=0,order_info='无',signal_price=float('nan'),order_close_chan=True,HedgeFlag:str="1",WaitReturn=False,OrderMemo="",**kw):
        '''
        报单,只支持限价下单,市价单以停板价报单,实际效果等效于市价单(部分交易所不支持市价单)
        symbol: str,合约代码
        kaiping: str,开平,kaiduo,kaikong,pingduo,pingkong,汉语拼音,表示 开多 开空 平多 平空
        lot: int,下单手数
        price:float,下单价格
        block:True,阻塞等待委托单结束
        n_price_tick: 排队价格偏离报单价多少跳不成交撤单,0不按价格偏离撤单
        che_time: 多少时间不成交则撤单,默认0不按时间撤单
        order_info:报单备注,例如是止损、止盈、策略1触发等
        signal_price:信号触发的价格位置
        order_close_chan:True,统计平仓单的盈亏
        '''
        if not self.check_instrument(symbol): 
            return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,"lot":lot,"price":price, "last_msg":f"合约码不存在,合约码:{symbol}" ,
                    "quote_volume":float('nan'),"order_id":[],"trades":[],"order_wrong":True,"signal_price":signal_price,"order_info":order_info,"quote":{},"position":{}}
        night_time = kw["night_time"] if "night_time" in kw else None  #夜盘收盘时间
        close_minutes = kw["close_minutes"] if "close_minutes" in kw else 0 #临近夜盘收盘多少分钟停止交易并撤单
        open_min_volume = kw['open_min_volume'] if 'open_min_volume' in kw else 1 #开仓最小手数
        order_wrong = False #是否错单
        quote = self.get_quote(symbol)
        #symbol = quote["InstrumentID"] #合约代码  
        #ExchangeID = quote["ExchangeID"] #交易所代码
        position = self.get_position(symbol)
        instrument_property = self.get_symbol_info(symbol)
        ExchangeID = instrument_property["ExchangeID"]
        PriceTick = instrument_property["PriceTick"]
        VolumeMultiple = instrument_property["VolumeMultiple"]
        ctp_time = (quote["ctp_datetime"]+timedelta(minutes=close_minutes)).time()
        equal_time =  quote["ctp_datetime"].hour == datetime.now().hour #行情和本地时间同步确保夜盘品种在白盘已更新
        trading_night_time = isinstance(night_time,time) and close_minutes > 0
        if trading_night_time : #只监控夜盘,白盘跨15点会结算
            if ctp_time >= night_time > time(20) or time(8) > ctp_time >= night_time :
                e = f"{datetime.now()} -open_close临近收盘停止交易,收盘时间:{night_time}\n"
                self.logs_txt(e,self._logfile,_print=False)
                return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,
                    "last_msg":f"临近收盘停止交易" ,
                    "order_id":[],"order_wrong":True if equal_time else False,"signal_price":signal_price,"order_info":order_info,
                    "quote":quote,"position":position}
        advanced = None
        quote_volume = quote["AskVolume1"] if kaiping in ["kaiduo","pingkong"] else quote["BidVolume1"] #盘口挂单量
        if not price or price == '对手价':
            if quote["AskPrice1"] <= quote["UpperLimitPrice"]:
                price_buy = quote["AskPrice1"]
            else: price_buy = quote["LastPrice"] #停板时以最新价报单
            if quote["BidPrice1"] >= quote["LowerLimitPrice"]:
                price_sell = quote["BidPrice1"]
            else: price_sell = quote["LastPrice"]
        elif price == '停板价':
            price_buy = quote["UpperLimitPrice"]
            price_sell = quote["LowerLimitPrice"]
        elif price == '排队价':
            if quote["BidPrice1"] >= quote["LowerLimitPrice"]:
                price_buy = quote["BidPrice1"]
            else: price_buy = quote["LastPrice"] #停板时以最新价报单
            if quote["AskPrice1"] <= quote["UpperLimitPrice"]:
                price_sell = quote["AskPrice1"]
            else: price_sell = quote["LastPrice"]
            advanced = None
        elif price == price :
            if price >= quote["UpperLimitPrice"] :
                price_buy = price_sell = quote["UpperLimitPrice"] #超出停板时以停板价报单
            elif price <= quote["LowerLimitPrice"]:
                price_buy = price_sell = quote["LowerLimitPrice"] #超出停板时以停板价报单
            else: price_buy = price_sell = price #其他限定价
            if price_buy < quote["AskPrice1"] and kaiping in ["kaiduo","pingkong"]:advanced = None
            if price_sell > quote["BidPrice1"] and kaiping in ["pingduo","kaikong"]:advanced = None
        if price != price or not isinstance(lot,int) or n_price_tick < 0 or che_time < 0:
            e = f"{datetime.now()} -open_close下单价格或数量或撤单跳数或撤单时间不合法,下单价格:{price},下单手数:{lot},撤单跳数:{n_price_tick},撤单时间:{che_time}\n"
            self.logs_txt(e,self._logfile,_print=False)
            return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,"lot":lot,"price":price,
                    "last_msg":f"下单价格或数量或撤单跳数或撤单时间不合法,下单价格:{price},下单手数:{lot},撤单跳数:{n_price_tick},撤单时间:{che_time}" ,
                    "quote_volume":quote_volume,"order_id":[],"trades":[],"order_wrong":True,"signal_price":signal_price,"order_info":order_info,"quote":quote,"position":position}
        lot = int(lot)
        lot = min(lot, instrument_property["MaxLimitOrderVolume"])
        if (price_buy/PriceTick - int(price_buy/PriceTick) != 0 or price_sell/PriceTick - int(price_sell/PriceTick) != 0 or
            lot <= 0 or lot < instrument_property["MinLimitOrderVolume"] or lot < open_min_volume): 
            #以持仓数量平仓时平仓数量为0可能是服务器故障持仓未更新(也可能其他程序超额平仓,或清仓代码正常所需非本策略bug),等待更新后可继续下单,其他情况下的报价和手数错误应退出交易
            if not (lot == position["pos_long"] == 0 and kaiping == 'pingduo' or lot == position["pos_short"] == 0 and kaiping=='pingkong'): 
                order_wrong = True 
            e = f"{datetime.now()} -open_close报单错误,下单价格:{price},下单手数:{lot},最大限价单:{instrument_property['MaxLimitOrderVolume']},最小限价单:{instrument_property['MinLimitOrderVolume']},最小开仓单:{open_min_volume},价格最小跳:{PriceTick}\n"
            self.logs_txt(e,self._logfile,_print=False)
            return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,"lot":lot,"price":price_buy if kaiping in ["kaiduo","pingkong"] else price_sell,
                    "last_msg":f"报单错误,下单价格:{price},下单手数:{lot},最大限价单:{instrument_property['MaxLimitOrderVolume']},最小限价单:{instrument_property['MinLimitOrderVolume']},最小开仓单:{open_min_volume},价格最小跳:{PriceTick}" ,
                    "quote_volume":quote_volume,"order_id":[],"trades":[],"order_wrong":order_wrong,"signal_price":signal_price,"order_info":order_info,"quote":quote,"position":position}
    
        shoushu = 0 #已成交手数  
        junjia = 0.0 #成交均价 
        che_count, day_order = 0, 0 #报撤单次数
        ping_jin,ping_zuo,order = None,None,None  #委托单对象
        last_msg, order_id, trades = "", [], []  #委托单状态信息和单号
        if kaiping== 'pingduo': #交易方向为平多
            #可能服务器故障持仓未更新,等待更新
            if not position["pos_long"] or position["open_price_long"] != position["open_price_long"]:
                e = f"{datetime.now()} -open_close多头持仓错误,多头持仓手数{position['pos_long']},多头持仓价格{position['open_price_long']}\n"
                self.logs_txt(e,self._logfile,_print=False)
                return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,"lot":lot,"price":price_buy if kaiping in ["kaiduo","pingkong"] else price_sell,
                        "last_msg":f"多头持仓错误,多头持仓手数{position['pos_long']},多头持仓价格{position['open_price_long']}" ,
                        "quote_volume":quote_volume, "order_id":[],"trades":[],"order_wrong":order_wrong,"signal_price":signal_price,"order_info":order_info,"quote":quote,"position":position}
            if 0 < lot <= position["pos_long_today"] : #小于等于今仓，平今
                ping_jin=self.insert_order(ExchangeID,symbol,'Sell','CloseToday', lot, price_sell,OrderMemo=OrderMemo)
                if isinstance(ping_jin,str): last_msg += ping_jin
            elif 0 < position["pos_long_today"] < lot <= position["pos_long"]:  #优先平今,剩余仓位平昨   
                ping_zuo=self.insert_order(ExchangeID,symbol,'Sell','Close',lot-position["pos_long_today"],price_sell,OrderMemo=OrderMemo) #先平昨再平今 
                ping_jin=self.insert_order(ExchangeID,symbol,'Sell','CloseToday',position["pos_long_today"],price_sell,OrderMemo=OrderMemo)
                if isinstance(ping_zuo,str): last_msg += ping_zuo
                if isinstance(ping_jin,str): last_msg += ping_jin
            elif 0 == position["pos_long_today"] < lot <= position["pos_long"]: #只有昨仓
                ping_zuo=self.insert_order(ExchangeID,symbol,'Sell','Close',lot,price_sell,OrderMemo=OrderMemo) 
                if isinstance(ping_zuo,str): last_msg += ping_zuo
            else: last_msg += "平仓数量不足"
        elif kaiping=='pingkong': #交易方向为平空
            if not position["pos_short"] or position["open_price_short"] != position["open_price_short"]:
                e = f"{datetime.now()} -open_close空头持仓错误,空头持仓手数{position['pos_short']},空头持仓价格{position['open_price_short']}\n"
                self.logs_txt(e,self._logfile,_print=False)
                return {"shoushu":0, "junjia":float("nan"), "che_count":0, "day_order":0, "symbol":symbol, "kaiping":kaiping,"lot":lot,"price":price_buy if kaiping in ["kaiduo","pingkong"] else price_sell,
                        "last_msg":f'空头持仓错误,空头持仓手数{position["pos_short"]},空头持仓价格{position["open_price_short"]}' ,
                        "quote_volume":quote_volume, "order_id":[],"trades":[],"order_wrong":order_wrong,"signal_price":signal_price,"order_info":order_info,"quote":quote,"position":position}
            if 0 < lot <= position["pos_short_today"] : #小于等于今仓，平今
                ping_jin=self.insert_order(ExchangeID,symbol,'Buy','CloseToday',lot,price_buy,OrderMemo=OrderMemo)    
                if isinstance(ping_jin,str): last_msg += ping_jin
            elif 0 < position["pos_short_today"] < lot <= position["pos_short"]:      
                ping_zuo=self.insert_order(ExchangeID,symbol,'Buy','Close',lot-position["pos_short_today"],price_buy,OrderMemo=OrderMemo)      
                ping_jin=self.insert_order(ExchangeID,symbol,'Buy','CloseToday',position["pos_short_today"],price_buy,OrderMemo=OrderMemo)
                if isinstance(ping_zuo,str): last_msg += ping_zuo
                if isinstance(ping_jin,str): last_msg += ping_jin
            elif 0 == position["pos_short_today"] < lot <= position["pos_short"]:      
                ping_zuo=self.insert_order(ExchangeID,symbol,'Buy','Close',lot,price_buy,OrderMemo=OrderMemo)
                if isinstance(ping_zuo,str): last_msg += ping_zuo
            else: last_msg += "平仓数量不足"
                
        elif kaiping== 'kaiduo': #交易方向为开多
            order = self.insert_order(ExchangeID,symbol,'Buy','Open',lot,price_buy,OrderMemo=OrderMemo)   
            if isinstance(order,str): last_msg += order
        elif kaiping=='kaikong': #交易方向为开空
            order = self.insert_order(ExchangeID,symbol,"Sell","Open",lot,price_sell,OrderMemo=OrderMemo)
            if isinstance(order,str): last_msg += order
        else: last_msg += f"交易方向{kaiping}只支持:kaiduo、kaikong、pingduo、pingkong"
        t = datetime.now().timestamp() #时间起点
        finished = ['0','4' ,'5' ]
        update_time = []
        local_timestamp = quote["local_timestamp"]
        if isinstance(ping_zuo,Order): #报单成功发向交易所
            orderid = ping_zuo["order_id"]
            order_id.append(orderid)
            last_price = quote.LastPrice
            while True:
                trades_list = self.get_trade_of_order(orderid,_print=False)
                trades.extend(trades_list)
                ping_zuo = self.get_id_order(order_id=orderid) #收到成交回报后再查询一次委托单更新,避免已全部成交的单触发误撤
                quote_volume = quote["AskVolume1"] if kaiping in ["kaiduo","pingkong"] else quote["BidVolume1"]
                ctp_time = (quote["ctp_datetime"]+timedelta(minutes=close_minutes)).time()
                if ping_zuo["OrderStatus"] not in finished or ping_zuo["VolumeTraded"] != sum( trade["Volume"] for trade in trades_list): #平昨单是否完成
                    if not advanced and not block :break #当日有效单，且无需等待是否完成
                    #等待che_time秒还不成交撤单，或者价格偏离委托价n_price_tick不成交撤单，适用于advanced=None的情况
                    if ping_zuo["OrderStatus"] not in finished and not advanced and (int(che_time)>0 and datetime.now().timestamp() - t >= che_time or int(n_price_tick)>0 and 
                    ((kaiping in ["kaiduo","pingkong"] and quote["BidPrice1"] >= ping_zuo["LimitPrice"]+PriceTick*n_price_tick) or
                    (kaiping in ["kaikong","pingduo"] and quote["AskPrice1"] <= ping_zuo["LimitPrice"]-PriceTick*n_price_tick)) 
                    and last_price != quote.LastPrice == quote.LastPrice 
                        or (trading_night_time and (ctp_time >= night_time > time(20) or time(8) > ctp_time >= night_time))
                    ):
                        self.cancel_order(orderid)  #等待撤单完成，防止重复撤单
                        last_price = quote.LastPrice
                else: break
            day_order += 1 #报单数加1
            volume = ping_zuo["VolumeTraded"] #成交手数
            if ping_zuo["VolumeTotal"] > 0: che_count += 1 #撤单次数增加
            last_msg += ping_zuo["StatusMsg"]
            if volume > 0: #有成交
                shoushu += volume #计算已成交手数
                for trade in trades_list:
                    junjia += trade["Volume"]*trade["Price"]
                    update_time.append(trade["local_timestamp"])

        if isinstance(ping_jin,Order): #报单成功发向交易所
            orderid = ping_jin["order_id"]
            order_id.append(orderid)
            last_price = quote.LastPrice
            while True:
                trades_list = self.get_trade_of_order(orderid,_print=False)
                trades.extend(trades_list)
                ping_jin = self.get_id_order(order_id=orderid) #收到成交回报后再查询一次委托单更新,避免已全部成交的单触发误撤
                quote_volume = quote["AskVolume1"] if kaiping in ["kaiduo","pingkong"] else quote["BidVolume1"]
                ctp_time = (quote["ctp_datetime"]+timedelta(minutes=close_minutes)).time()
                if ping_jin["OrderStatus"] not in finished or ping_jin["VolumeTraded"] != sum( trade["Volume"] for trade in trades_list): #平昨单是否完成
                    if not advanced and not block :break #当日有效单，且无需等待是否完成
                    #等待che_time秒还不成交撤单，或者价格偏离委托价n_price_tick不成交撤单，适用于advanced=None的情况
                    if ping_jin["OrderStatus"] not in finished and not advanced and (int(che_time)>0 and datetime.now().timestamp() - t >= che_time or int(n_price_tick)>0 and 
                    ((kaiping in ["kaiduo","pingkong"] and quote["BidPrice1"] >= ping_jin["LimitPrice"]+PriceTick*n_price_tick) or
                    (kaiping in ["kaikong","pingduo"] and quote["AskPrice1"] <= ping_jin["LimitPrice"]-PriceTick*n_price_tick)) 
                    and last_price != quote.LastPrice == quote.LastPrice 
                        or (trading_night_time and (ctp_time >= night_time > time(20) or time(8) > ctp_time >= night_time))
                    ):
                        self.cancel_order(orderid)  #等待撤单完成，防止重复撤单
                        last_price = quote.LastPrice
                else: break
            day_order += 1 #报单数加1
            volume = ping_jin["VolumeTraded"] #成交手数
            if ping_jin["VolumeTotal"] > 0: che_count += 1 #撤单次数增加
            last_msg += ping_jin["StatusMsg"]
            if volume > 0: #有成交
                shoushu += volume #计算已成交手数
                for trade in trades_list:
                    junjia += trade["Volume"]*trade["Price"]
                    update_time.append(trade["local_timestamp"])

        if isinstance(order,Order): #报单成功发向交易所
            orderid = order["order_id"]
            order_id.append(orderid)
            last_price = quote.LastPrice
            while True:
                trades_list = self.get_trade_of_order(orderid,_print=False)
                trades.extend(trades_list)
                order = self.get_id_order(order_id=orderid) #收到成交回报后再查询一次委托单更新,避免已全部成交的单触发误撤
                quote_volume = quote["AskVolume1"] if kaiping in ["kaiduo","pingkong"] else quote["BidVolume1"]
                ctp_time = (quote["ctp_datetime"]+timedelta(minutes=close_minutes)).time()
                if order["OrderStatus"] not in finished or order["VolumeTraded"] != sum( trade["Volume"] for trade in trades_list): #平昨单是否完成
                    if  not block :break #当日有效单，且无需等待是否完成
                    #等待che_time秒还不成交撤单，或者价格偏离委托价n_price_tick不成交撤单，适用于advanced=None的情况
                    if order["OrderStatus"] not in finished and  (int(che_time)>0 and datetime.now().timestamp() - t >= che_time or int(n_price_tick)>0 and 
                    ((kaiping in ["kaiduo","pingkong"] and quote["BidPrice1"] >= order["LimitPrice"]+PriceTick*n_price_tick) or
                    (kaiping in ["kaikong","pingduo"] and quote["AskPrice1"] <= order["LimitPrice"]-PriceTick*n_price_tick)) 
                    and last_price != quote.LastPrice == quote.LastPrice 
                        or (trading_night_time and (ctp_time >= night_time > time(20) or time(8) > ctp_time >= night_time))
                    ):
                        self.cancel_order(orderid)  #等待撤单完成，防止重复撤单
                        last_price = quote.LastPrice
                else: break
            day_order += 1 #报单数加1
            volume = order["VolumeTraded"] #成交手数
            if order["VolumeTotal"] > 0: che_count += 1 #撤单次数增加
            last_msg += order["StatusMsg"]
            if volume > 0: #有成交
                shoushu += volume #计算已成交手数
                for trade in trades_list:
                    junjia += trade["Volume"]*trade["Price"]
                    update_time.append(trade["local_timestamp"])
        if shoushu: 
            junjia = junjia/shoushu #计算成交均价
            #junjia = round(junjia/shoushu,quote.price_decs) #保留和报价同样小数位
            for trade in trades_list:
                trade_id = f'{trade["ExchangeID"]}_{trade["OrderSysID"]}_{trade["TradeID"]}'
                while trade_id not in self.trades: tm.sleep(0.001)
                
        else: #错单原因
            junjia = float('nan')
            #集合竞价,未到开盘时间等待60秒
            if "拒绝" in last_msg and ("竞价" in last_msg or "竟价" in last_msg or "交易时间" in last_msg): tm.sleep(60)
            elif ("拒绝" in last_msg or "限制" in last_msg or "不足" in last_msg or "超过" in last_msg or "不支持" in last_msg or "低于" in last_msg
                  or "权限" in last_msg or "平仓" in last_msg or "开仓" in last_msg or "禁止" in last_msg or "状态" in last_msg or "操作" in last_msg
                  or "交易" in last_msg or "产品" in last_msg or "CTP" in last_msg or "开户" in last_msg or "不允许" in last_msg or "报单" in last_msg
                   or "交易所" in last_msg or "合约" in last_msg or "确认" in last_msg or "错误" in last_msg or "有误" in last_msg):
                order_wrong = True
        e = f"{datetime.now()} -open_close下单完成,合约:{symbol},交易方向{kaiping},成交手数:{shoushu},报单价:{price},成交均价:{junjia},委托单信息:{last_msg},备注:{order_info}\n"
        self.logs_txt(e,self._logfile,_print=False)    
        return {"shoushu":shoushu, "junjia":junjia, "che_count":che_count, "day_order":day_order, "symbol":symbol, "kaiping":kaiping, "lot":lot, "price":price_buy if kaiping in ["kaiduo","pingkong"] else price_sell, "last_msg":last_msg, 
                "quote_volume":quote_volume, "order_id":order_id,"trades":trades,"order_wrong":order_wrong,"signal_price":signal_price,"order_info":order_info,"quote":quote,"position":position} #返回成交手数、成交均价,和主动撤单次数，若无成交则均价为nan值
        
    def _rtn_thread(self):
        for i in self._rtn_queue:
            if i == "exception":  
                return
            if isinstance(i,dict):
                t = tm.time()
                if "instruments" in i:
                    self.instruments = i["instruments"] #全部合约码
                elif "pos" in i:
                    for s in i["pos"]:
                        if s not in self.positions: self.positions[s] = Position( )
                        self.positions[s].update(i["pos"][s])
                        self.positions[s].update({"local_timestamp":t})
                elif "quote" in i:
                    InstrumentID = i["quote"]["InstrumentID"]
                    if InstrumentID in self.unsubscribe_instruments: continue
                    if InstrumentID not in self.quote: 
                        self.quote[InstrumentID] = Quote( )
                    #if self.quote[InstrumentID]["TradingDay"] != self.TradingDay: self.quote[InstrumentID]["TradingDay"] = self.TradingDay
                    #try:
                        # 确保毫秒部分为3位数
                    #    full_time_str = f'{i["quote"]["ActionDay"]} {i["quote"]["UpdateTime"]}.{i["quote"]["UpdateMillisec"]:03d}'
                    #    tick_time = datetime.strptime(full_time_str, "%Y%m%d %H:%M:%S.%f")
                    #except ValueError:
                    #    tick_time = datetime.strptime(f'{i["quote"]["ActionDay"]} {i["quote"]["UpdateTime"]}.{i["quote"]["UpdateMillisec"]}')
                    if i["quote"]["ActionDay"]:
                        dt = datetime.strptime(f'{i["quote"]["ActionDay"]} {i["quote"]["UpdateTime"]}.{i["quote"]["UpdateMillisec"]:03d}', "%Y%m%d %H:%M:%S.%f")
                    else: dt = datetime.now()
                    if dt == self.quote[InstrumentID].ctp_datetime: dt += timedelta(milliseconds=500) #郑商所毫秒值均为0
                    dt_timestamp = dt.timestamp()
                    trading_day_date = datetime.strptime(i["quote"]["TradingDay"], "%Y%m%d").date()
                    self.quote[InstrumentID].update(i["quote"])
                    self.quote[i["quote"]["InstrumentID"]].update({"local_timestamp":t,"ctp_timestamp":dt_timestamp,"ctp_datetime":dt,"trading_day":trading_day_date})
                    if self._save_tick:
                        if i["quote"]["InstrumentID"] in self._save_tick_symbols:
                            self._save_tick_queue.put(self.quote[i["quote"]["InstrumentID"]].to_dict())
                    #if "notify" not in self.quote[i["quote"]["InstrumentID"]]: self.quote[i["quote"]["InstrumentID"]]["notify"] = zhuchannel.ThreadChan(last_only=True)
                    #self.quote[i["quote"]["InstrumentID"]]["notify"].send(True)
                elif "account" in i:
                    self.account.update(i["account"])
                    self.account.update({"local_timestamp":t})
                elif "order" in i:
                    for order_id in i["order"]:
                        if order_id not in self.orders: self.orders[order_id] = Order( )
                        self.orders[order_id].update(i["order"][order_id]) 
                        self.orders[order_id].update({"local_timestamp":t})
                elif "trade" in i:
                    for trade_id in i["trade"]:
                        if trade_id not in self.trades: self.trades[trade_id] = Trade( )
                        self.trades[trade_id].update(i["trade"][trade_id]) 
                        self.trades[trade_id].update({"local_timestamp":t})
                elif "orderaction" in i:
                    for order_id in i["orderaction"]:
                        if order_id not in self.orders: self.orders[order_id] = Order( )
                        self.orders[order_id].update(i["order"][order_id]) 
                        self.orders[order_id].update({"local_timestamp":t})
    def _ans_thread(self):
        for i in self._ans_queue:
            if not isinstance(i,dict): 
                print('CTP连接中断,重新登录')
                continue
            if "instruments" in i:
                self.instruments = i["instruments"]
            elif "pos" in i:
                t = tm.time()
                for s in i["pos"]:
                    if s not in self.positions: self.positions[s] = Position(**i["pos"][s])
                    else: self.positions[s].update(i["pos"][s])
                    self.positions[s].update({"local_timestamp":t})
    def check_instrument(self,InstrumentID:str):
        if InstrumentID in self.instruments: return True
        else: 
            e = f"{datetime.now()} -合约{InstrumentID}不存在\n"
            self.logs_txt(e,self._logfile,_print=True)
            #raise Exception(f"合约{InstrumentID}不存在")
            return False
     
    def _sendReq(self,kw):
        '''
        不需要向ctp查询的直接收到的是结果,需要向ctp查询的首先收到的是发送成败
        r[ReqID]True和False表示向CTP发送成功或失败,r["return"]表示r[ReqID]为最终结果
        '''
        #等待CTP准备就绪
        while not self._Reqqueue.empty(): tm.sleep(0.000000001) 
        self._ReqID += 1
        ReqID = self._ReqID
        self._Reqqueue.put({**kw,"ReqID":ReqID})
        while True:
            r = self._ans_queue.get()
            if not isinstance(r,dict): continue  #断线重连,非数据反馈
            if ReqID in r and "return" in r and r["return"]: break #查询完成,得到最终结果
            elif ReqID not in r: self._ans_queue.put_nowait(r) #非本次查询
        return {"ReqID":ReqID,"ret":r[ReqID]}
    
    def subscribe_quote(self,InstrumentID:str,wait_return=False):
        '''只负责行情订阅,最新行情从quote中取'''
        if isinstance(InstrumentID,str): instrument_ids = [InstrumentID]
        elif isinstance(InstrumentID,list): instrument_ids = InstrumentID
        for s in instrument_ids:
            if not self.check_instrument(s): return
            if not s: 
                e = f"{datetime.now()} -合约码为空,订阅行情失败\n"
                self.logs_txt(e,self._logfile,_print=True)
                return
        #for s in InstrumentID:
        #    self.reqry_market_data(s)
        '''等待行情返回'''
        ss = [s for s in InstrumentID if s not in self.quote]
        if ss: 
            self._Reqqueue.put_nowait({"Subscribe":ss})
            for s in ss:
                if s in self.unsubscribe_instruments: self.unsubscribe_instruments.remove(s)
        
    def unsubscribe_quote(self,InstrumentID:str,wait_return=False):
        if isinstance(InstrumentID,str): instrument_ids = [InstrumentID]
        elif isinstance(InstrumentID,list): instrument_ids = InstrumentID
        for s in instrument_ids:
            if not self.check_instrument(s): return
            if not s: 
                e = f"{datetime.now()} -合约码为空,退订行情失败\n"
                self.logs_txt(e,self._logfile,_print=True)
                return
        ss = [s for s in instrument_ids if s in self.quote and s not in self.positions ] #有行情且无持仓才能退订
        if ss:
            self._Reqqueue.put_nowait({"UnSubscribe":ss})
            for s in ss: 
                if s in self.quote: self.quote.pop(s)
                if s not in self.unsubscribe_instruments: self.unsubscribe_instruments.add(s)
                    

    def reqry_market_data(self,InstrumentID:str,wait_return=False):
        return self.get_quote(InstrumentID,wait_return,subscribe=False)

    def get_kline(self,InstrumentID: str, period: str, kline_count:int, save_instrument_id: str = "") -> Optional[MutableKlineHolder]:
        '''
        Args:
            InstrumentID: 合约代码
            period: 周期,单位为s、m、h、d、w、M、y,表示秒、分、时、日、周、月、年,一周表示7天,一月表示30天,一年表示365天
            kline_count: k线数量,当本地tick数据不足以生成kline_count数量的K线时,则只返回可生成的数量,并在后续tick推送中逐渐更新到kline_count数量
            save_instrument_id:保存tick数据的名称,默认为合约码
        Return:
            返回MutableKlineHolder对象,通过data属性取值为polars.DataFrame
            包含基础列"InstrumentID","open","high","low","close","Volume","Turnover","OpenInterest"以及Quote对象字段
            "period_start" K线开始时间点, "period_end" K线结束时间点
            "period" K线周期
        用法:
            kline = get_kline("rb2110","1d",10)
            print(kline.data) #打印全部K线数据
            print(kline.tail(10)) #打印最新10根K线数据
        '''
        if not self.check_instrument(InstrumentID): return
        if InstrumentID not in self.quote: self.get_quote(InstrumentID)
        self._save_tick = True
        self._save_tick_symbols[InstrumentID] = save_instrument_id
        kline = self.CTP_data_processor.get_kline_holder(InstrumentID , period , kline_count)
        return kline
    
    def get_tick(self,InstrumentID: str, save_instrument_id: str = "") -> Optional[MutableTickHolder]:
        '''
        Args:
            InstrumentID: 合约代码
            save_instrument_id:保存tick数据的名称,默认为合约码。如果保存主连数据,则所有主力合约(01/05/09)都存储相同的名称
        Return:
            返回MutableTickHolder对象,通过data属性取值为polars.DataFrame
            包含列参见Quote对象字段
            
        用法:
            tick = get_tick("rb2110")
            print(tick.data) #打印全部Tick数据
            print(tick.tail(10)) #打印最新10根Tick数据
        '''
        if not self.check_instrument(InstrumentID): return
        if InstrumentID not in self.quote: self.get_quote(InstrumentID)
        self._save_tick = True
        self._save_tick_symbols[InstrumentID] = save_instrument_id
        tick = self.CTP_data_processor.get_tick_holder(InstrumentID )
        return tick

    def _update_tick_kline(self):
        for i in self._save_tick_queue:
            if i == "exception":  
                return
            self.CTP_data_processor.process_snapshot(snapshot = i, save_instrument_id = self._save_tick_symbols[i["InstrumentID"]])

    async def OpenClose(self,symbol:str, kaiping: str = '', lot: int = 0, price: float = None,block: bool = True, 
                        n_price_tick: int = 1, che_time: int = 0, order_info: str = '无', signal_price: float = float('nan'),
                        order_close_chan: bool = True, HedgeFlag: str = "1", WaitReturn: bool = False,OrderMemo="",**kw):
        '''open_close的协程版'''
        # 关键：用partial绑定open_close的关键字参数
        bound_func = partial(self.open_close, symbol=symbol, kaiping = kaiping, lot = lot, price = price, block = block,
                            n_price_tick = n_price_tick, che_time = che_time, order_info = order_info, signal_price = signal_price,
                            order_close_chan = order_close_chan, HedgeFlag = HedgeFlag, WaitReturn = WaitReturn, OrderMemo = OrderMemo,**kw)  # 先把参数绑定到函数上
        
        loop = asyncio.get_running_loop()
        # 此时run_in_executor只需传绑定后的函数，无需额外参数
        result = await loop.run_in_executor(self.executor, bound_func)
        return result
    
    # 通用异步协程：调用阻塞函数（通过线程池）
    async def async_wrapper(self,func,*args,**kws ):
        '''将阻塞函数提交进线程池(若线程池已满需排队等待直到有线程执行结束),以支持协程中使用,事件循环工作在某单个线程中(可以不属于self.executor)'''
        # 关键：用partial绑定open_close的关键字参数
        bound_func = partial(func,*args,**kws)  # 先把参数绑定到函数上
        loop = asyncio.get_running_loop()
        # 此时run_in_executor只需传绑定后的函数，无需额外参数
        result = await loop.run_in_executor(self.executor, bound_func)
        return result
    
    def auto_pos_thread(self,symbol,return_queue:Optional[queue.Queue]=None):
        '''
        智能调仓任务,自动下单,并随排队价相对报单价变动自动撤单、追单,直到持仓达到设定仓位,若是下单出错(如保证金不足)则退出调仓任务
        Args:
            symbol: 合约代码
            return_queue: 下单结果返回值队列
        '''
        if not self.check_instrument(symbol): return
        if symbol not in self._pos_queue: self._pos_queue[symbol] = queue.Queue()
        else: 
            e = f"{datetime.now()} -合约{symbol}的调仓任务已存在\n"
            self.logs_txt(e,self._logfile,_print=True)
            return
        quote = self.get_quote(symbol)
        position = self.get_position(symbol)
        instrument_property = self.get_symbol_info(symbol)
        PriceTick = instrument_property["PriceTick"]
        while True:
            if symbol not in self._pos_queue: return
            #{"pos_long":int,"pos_short":int,"price":"对手价"/"排队价/超价","open_min_volume":int,"kw":{}}
            pos = self._pos_queue[symbol].get()
            
            if isinstance(pos["pos_long"],int):
                while position.pos_long != pos["pos_long"] :
                    if position.pos_long < pos["pos_long"] : 
                        kaiping = "kaiduo"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.AskPrice1
                        elif pos['price'] == "排队价": price = quote.BidPrice1
                        elif pos['price'] == "超价": price = quote.BidPrice1 + 1 * PriceTick

                    else: 
                        kaiping = "pingduo"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.BidPrice1
                        elif pos['price'] == "排队价": price = quote.AskPrice1
                        elif pos['price'] == "超价": price = quote.AskPrice1 - 1 * PriceTick
                    lot = abs(position.pos_long - pos["pos_long"] )
                    if kaiping == "kaiduo" and lot < pos["open_min_volume"]:
                        e = f"{datetime.now()} -开仓数量:{lot},小于最低开仓数量:{pos['open_min_volume']}\n"
                        self.logs_txt(e,self._logfile,_print=True)
                        break
                    elif lot:
                        r = self.open_close(symbol,kaiping,lot,price,**pos["kw"])
                        if isinstance(return_queue,queue.Queue): return_queue.put(r)
                        if not r["shoushu"] and r["order_wrong"]:
                            e = f"{datetime.now()} -下单出错,返回值:{r}\n"
                            self.logs_txt(e,self._logfile,_print=True)
                            self._pos_queue.pop(symbol)
                            return
               
            if isinstance(pos["pos_short"],int):
                while position.pos_short != pos["pos_short"] :
                    if position.pos_short < pos["pos_short"] : 
                        kaiping = "kaikong"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.BidPrice1
                        elif pos['price'] == "排队价": price = quote.AskPrice1
                        elif pos['price'] == "超价": price = quote.AskPrice1 - 1 * PriceTick
                    else: 
                        kaiping = "pingkong"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.AskPrice1
                        elif pos['price'] == "排队价": price = quote.BidPrice1
                        elif pos['price'] == "超价": price = quote.BidPrice1 + 1 * PriceTick
                    lot = abs(position.pos_short - pos["pos_short"] )
                    if kaiping == "kaikong" and lot < pos["open_min_volume"]:
                        e = f"{datetime.now()} -开仓数量:{lot},小于最低开仓数量:{pos['open_min_volume']}\n"
                        self.logs_txt(e,self._logfile,_print=True)
                        break    
                    elif lot:
                        r = self.open_close(symbol,kaiping,lot,price,**pos["kw"])
                        if isinstance(return_queue,queue.Queue): return_queue.put(r)
                        if not r["shoushu"] and r["order_wrong"]:
                            e = f"{datetime.now()} -下单出错,返回值:{r}\n"
                            self.logs_txt(e,self._logfile,_print=True)
                            self._pos_queue.pop(symbol)
                            return
                
    async def auto_pos_async(self,symbol,return_queue:Optional[asyncio.Queue]=None):
        '''
        智能调仓任务,自动下单,并随排队价相对报单价变动自动撤单、追单,直到持仓达到设定仓位,若是下单出错(如保证金不足)则退出调仓任务
        Args:
            symbol: 合约代码
            return_queue: 下单结果返回值队列
        '''
        if not self.check_instrument(symbol): return
        if symbol not in self._pos_queue: self._pos_queue[symbol] = asyncio.Queue()
        else: 
            e = f"{datetime.now()} -合约{symbol}的调仓任务已存在\n"
            self.logs_txt(e,self._logfile,_print=True)
            return
        loop = asyncio.get_running_loop()
        quote = await loop.create_task(self.async_wrapper(self.get_quote,symbol))
        position = await loop.create_task(self.async_wrapper(self.get_position,symbol))
        instrument_property = self.get_symbol_info(symbol)
        PriceTick = instrument_property["PriceTick"]
        while True:
            if symbol not in self._pos_queue: return
            #{"pos_long":int,"pos_short":int,"price":"对手价"/"排队价","open_min_volume":int,"kw":{}}
            pos = await self._pos_queue[symbol].get()
            if isinstance(pos["pos_long"],int):
                while position.pos_long != pos["pos_long"] :
                    if position.pos_long < pos["pos_long"] : 
                        kaiping = "kaiduo"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.AskPrice1
                        elif pos['price'] == "排队价": price = quote.BidPrice1
                        elif pos['price'] == "超价": price = quote.BidPrice1 + 1 * PriceTick
                    else: 
                        kaiping = "pingduo"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.BidPrice1
                        elif pos['price'] == "排队价": price = quote.AskPrice1
                        elif pos['price'] == "超价": price = quote.AskPrice1 - 1 * PriceTick
                    lot = abs(position.pos_long - pos["pos_long"] )
                    if kaiping == "kaiduo" and lot < pos["open_min_volume"]:
                        e = f"{datetime.now()} -开仓数量:{lot},小于最低开仓数量:{pos['open_min_volume']}\n"
                        self.logs_txt(e,self._logfile,_print=True)
                        break
                    elif lot:
                        r = await loop.create_task(self.OpenClose(symbol,kaiping,lot,price,**pos["kw"]))
                        if isinstance(return_queue,asyncio.Queue): await return_queue.put(r)
                        if not r["shoushu"] and r["order_wrong"]:
                            e = f"{datetime.now()} -下单出错,返回值:{r}\n"
                            self.logs_txt(e,self._logfile,_print=True)
                            self._pos_queue.pop(symbol)
                            return
            if isinstance(pos["pos_short"],int):
                while position.pos_short != pos["pos_short"] :
                    if position.pos_short < pos["pos_short"] : 
                        kaiping = "kaikong"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.BidPrice1
                        elif pos['price'] == "排队价": price = quote.AskPrice1
                        elif pos['price'] == "超价": price = quote.AskPrice1 - 1 * PriceTick
                    else: 
                        kaiping = "pingkong"
                        if pos['price'] == "对手价" or pos['price'] is None: price = quote.AskPrice1
                        elif pos['price'] == "排队价": price = quote.BidPrice1
                        elif pos['price'] == "超价": price = quote.BidPrice1 + 1 * PriceTick
                    lot = abs(position.pos_short - pos["pos_short"] )
                    if kaiping == "kaikong" and lot < pos["open_min_volume"]:
                        e = f"{datetime.now()} -开仓数量:{lot},小于最低开仓数量:{pos['open_min_volume']}\n"
                        self.logs_txt(e,self._logfile,_print=True)
                        break    
                    elif lot:
                        r = await loop.create_task(self.OpenClose(symbol,kaiping,lot,price,**pos["kw"]))
                        if isinstance(return_queue,asyncio.Queue): await return_queue.put(r)
                        if not r["shoushu"] and r["order_wrong"]:
                            e = f"{datetime.now()} -下单出错,返回值:{r}\n"
                            self.logs_txt(e,self._logfile,_print=True)
                            self._pos_queue.pop(symbol)
                            return
                
    def set_pos_volume(self, symbol, pos_long:int=None, pos_short:int=None, price:Optional[str]=None, open_min_volume:int=1, **kw):
        '''
        设置目标仓位,由调仓任务auto_pos自动把持仓调整到目标数量,设置超价则以超排队价1跳下单
        用法:
            #设置多头目标仓位100手 
            set_pos_volume(rb2601, pos_long=100 )
            #设置多头目标仓位100手,以排队价下单 
            set_pos_volume(rb2601, pos_long=100, price = "排队价" )

            #设置空头目标仓位10手
            set_pos_volume(rb2601, pos_short=10)

            #同时设置多头目标仓位100手、空头目标仓位10手,auto_pos会先完成多头调仓再完成空头调仓
            set_pos_volume(rb2601, pos_long=100, pos_short=10)

            #设置多头目标仓位10手, 最低开仓数量4手(如菜粕交易所规定最低开仓4手)
            set_pos_volume(RM601, pos_long=10, open_min_volume=4)
        Args:
            symbol: 合约代码
            pos_long: 多仓数量
            pos_short: 空仓数量
            price: 下单价格,只支持"对手价"、"排队价"、"超价",None为对手价,不支持其他指定价避免传入错误参数频繁报撤单
            open_min_volume: 开仓最小数量,交易所最低开仓数量合约可设置
            **kw: 其他参数,若设置cancel=True,则取消调仓任务
        '''
        if not self.check_instrument(symbol): return
        if symbol in self._pos_queue: 
            if pos_long is None and pos_short is None: return
            if "cancel" in kw and kw["cancel"]:
                self._pos_queue.pop(symbol)
                return
            order_data = {"symbol":symbol, "pos_long":pos_long, "pos_short":pos_short, "price":price, "open_min_volume":open_min_volume, "kw":kw}
            self._pos_queue[symbol].put_nowait(order_data)
        else: 
            e = f"{datetime.now()} -合约{symbol}智能单任务已结束\n"
            self.logs_txt(e,self._logfile,_print=True)

    #日志            
    def logs_txt(self,e,logfile="",_print=True,sf=""):
        if not logfile:
            logfile = os.path.dirname(os.path.abspath(__file__))
        if _print: print(e)
        os.makedirs(logfile,exist_ok=True)
        ss = logfile+f"\\{date.today()}{sf}.txt"
        ff = open(ss,mode="a+",encoding='utf-8')
        ff.write(e)
        ff.write(f'\n{"-"*30}\n')
        ff.close()

    def trade_excel(self,new_df:polars.DataFrame,direction,filename, logfile = ""):
        '''保存成交记录为csv,查看文件时建议复制一份再查看,避免对文件改动引起后续写入出错'''
        if not logfile:
            logfile = os.path.dirname(os.path.abspath(__file__))
        #logfile = r'C:\TqLogs'  #指定目录
        os.makedirs(logfile,exist_ok=True)
        #"""
        file_path = fr'{logfile}\策略{direction}统计-{filename}.csv'
        include_header = not os.path.exists(file_path)
        try:
            with open(file_path, "a", encoding="utf-8") as f:
                new_df.write_csv( f, include_header=include_header )
        except Exception as e: #文件被占用或其他原因无法写入
            self.logs_txt(f"保存成交记录失败,原因:{e}",file_path)
            file_path = fr'{logfile}\策略{direction}统计-{filename}{datetime.today()}.csv'
            with open(file_path, "a", encoding="utf-8") as f:
                new_df.write_csv( f, include_header=include_header )

    def logs_excel(self,new_df:polars.DataFrame,filename,logfile=""):
        '''保存每一个策略的盈亏统计'''
        if not logfile:
            logfile = os.path.dirname(os.path.abspath(__file__))
        os.makedirs(logfile,exist_ok=True)
        file_path = fr"{logfile}\策略盈亏统计-{filename}.csv"
        include_header = not os.path.exists(file_path) 
        try:
            with open(file_path, "a", encoding="utf-8") as f:
                new_df.write_csv( f, include_header=include_header)
        except Exception as e: #文件被占用或其他原因无法写入
            self.logs_txt(f"保存策略盈亏统计失败,原因:{e}",file_path)
            file_path = fr"{logfile}\策略盈亏统计-{filename}{datetime.today()}.csv"
            with open(file_path, "a", encoding="utf-8") as f:
                new_df.write_csv( f,  include_header=include_header )

    def save_json(self,filehand,datadict):
        #保存本地文件
        filehand.seek(0) #
        filehand.truncate() #
        json.dump(datadict, filehand,indent=4) 
        filehand.flush() #

    #发邮件
    def send_email(self,sender,authorization,recver_list,title,content):
        try: 
            #sender = "2522729256@qq.com"  #发送方
            #authorization = "你的自己的SMTP授权码" #SMTP授权码，不是邮箱密码
            #recver_list = ["2522729256@qq.com"] #接收方列表
            #title = "算法通知"         #邮件标题
            if isinstance(recver_list,str): recver_list = [recver_list]
            message = MIMEText(content, "plain", "utf-8")  #"plain"文本格式  
            message['Subject'] = title                   #邮件标题
            message['From'] = sender                       #发件人
            message['To'] = ",".join(recver_list)         # 收件人列表
            #windows server 2012 服务器上的程序发送smtp邮件必须通过SSL协议端口发送
            #smtp = smtplib.SMTP("smtp.qq.com",465,timeout=3)
            smtp = smtplib.SMTP_SSL("smtp.qq.com",465,timeout=3) #实例化smtp服务器 163邮箱用465或994端口邮件服务商的加密465端口，否则用25端口
            smtp.login(sender, authorization)               #发件人登录
            smtp.sendmail(sender, recver_list,
                            message.as_string())     # as_string 对 message 的消息进行了封装
            smtp.quit()                  
        except: pass
        finally:
            smtp.close()
    # 发消息
    def send_msg(self,DingDing,WeChat,QQemail,chan:zhuchannel.ThreadChan,logfile,_print=False,sf="") :
        for content in chan:
            try:
                headers = {"content-type": "application/json;charset=utf-8"}
                if  DingDing: 
                    msg = {"msgtype": "text",
                        "text": {"content": "-{}\n{}".format(DingDing[0],content)}}
                    body = json.dumps(msg)
                    requests.post(DingDing[1], data=body, headers=headers,timeout=3 )
                if  WeChat: 
                    msg = {"msgtype": "text",
                        "text": {"content": "-{}\n{}".format(WeChat[0],content)}}
                    body = json.dumps(msg)
                    requests.post(WeChat[1], data=body, headers=headers,timeout=3 )
                if QQemail: self.send_email(*QQemail,content)
                #self.logs_txt(content,s)
                tm.sleep(4)
            except Exception: 
                e = traceback.format_exc()
                self.logs_txt(e,logfile,_print,sf)

    def send_message(self,DingChan:zhuchannel.ThreadChan,account,orderrisk={},r={},name="",e="",**kw):
        if isinstance(DingChan,zhuchannel.ThreadChan): 
            if r:
                position = r['position']
                content = "{}{}{}{}{}".format(
                                    f"{datetime.now()}-{e}\n策略-{name}\n{r['symbol']} {r['kaiping']},下单手数:{r['lot']},下单价格:{r['price']},成交手数:{r['shoushu']},成交均价:{r['junjia']},",
                                    f"{r['last_msg']},\n{'空头' if 'kong' in r['kaiping'] else '多头'}持仓:{position['pos_short'] if 'kong' in r['kaiping'] else position['pos_long']}\n",
                                    f"账户权益:{round(account['Balance'],2)},账户可用资金:{round(account['Available'],2)},账户保证金:{round(account['CurrMargin'],2)},持仓浮盈:",
                                    f"{round(account['float_profit'],2)},风险度:{round(account['risk_ratio'],2)},\n",
                                    f"\n报撤单统计:{orderrisk}")
            else:
                content = "{}{}{}{}".format(
                                    f"{datetime.now()}-{e}\n策略-{name} ",
                                    f"账户权益:{round(account['Balance'],2)},账户可用资金:{round(account['Available'],2)},账户保证金:{round(account['CurrMargin'],2)},持仓浮盈:",
                                    f"{round(account['float_profit'],2)},风险度:{round(account['risk_ratio'],2)},\n",
                                    f"报撤单统计:{orderrisk}")
            ln = len(content)
            lln = 2000 if "msg" not in  kw else kw["msg"]
            num = int(ln/lln) + 1 if ln/lln - int(ln/lln) > 0 else int(ln/lln)
            for i in range(1,num+1):
                DingChan.put_nowait(f"第{i}部分,共{num}部分\n\n" + content[(i-1)*lln:lln*i])






#!usr/bin/env python3
#-*- coding:utf-8 -*-
__author__ = 'zhuxueli'
from typing import Dict, Any, Optional, Iterable, List, Tuple, Iterator
import polars,os
import glob,re
from datetime import datetime,date,timedelta

nReason = {4097:'网络读失败',4098:'网络写失败',8193:'接收心跳超时',8194:'发送心跳失败',8195:'收到错误报文'}
HedgeFlag={'speculation':'1',   #投机
            'arbitrage':'2',    #套利
            'hedge':'3'         #套保
           }

class FullDictDataClass():
    """支持全部字典方法的可变数据类"""
    
    def __init__(self, data: Optional[Dict[str, Any]] = None):
        """从字典初始化属性"""
        if data: self.__dict__.update(data)
    
    # ----------------------
    # 核心：字典式访问基础方法
    # ----------------------
    def __getitem__(self, key: str) -> Any:
        """支持 obj[key] 访问"""
        if key in self.__dict__:
            return self.__dict__[key]
        raise KeyError(f"键不存在: {key}")
    
    def __setitem__(self, key: str, value: Any) -> None:
        """支持 obj[key] = value 修改"""
        self.__dict__[key] = value
    
    def __delitem__(self, key: str) -> None:
        """支持 del obj[key] 删除"""
        if key in self.__dict__:
            del self.__dict__[key]
        else:
            raise KeyError(f"键不存在: {key}")
    
    def __iter__(self) -> Iterator[str]:
        """默认迭代键（与字典行为一致）"""
        return iter(self.__dict__.keys())  # 迭代键

    # ----------------------
    # 实现字典的全部方法
    # ----------------------
    def keys(self) -> Iterable[str]:
        """返回所有键，同 dict.keys()"""
        return self.__dict__.keys()
    
    def values(self) -> Iterable[Any]:
        """返回所有值，同 dict.values()"""
        return self.__dict__.values()
    
    def items(self) -> Iterable[Tuple[str, Any]]:
        """返回所有键值对，同 dict.items()"""
        return self.__dict__.items()
    
    def get(self, key: str, default: Any = None) -> Any:
        """获取键值，不存在返回默认值，同 dict.get()"""
        return self.__dict__.get(key, default)
    
    def update(self, other: Dict[str, Any]) -> 'FullDictDataClass':
        """批量更新键值对，同 dict.update()"""
        self.__dict__.update(other)
        return self
    
    def pop(self, key: str, default: Any = None) -> Any:
        """删除并返回键值，同 dict.pop()"""
        return self.__dict__.pop(key, default)
    
    def popitem(self) -> Tuple[str, Any]:
        """删除并返回最后一个键值对，同 dict.popitem()"""
        return self.__dict__.popitem()
    
    def clear(self) -> None:
        """清空所有键值对，同 dict.clear()"""
        self.__dict__.clear()
    
    def copy(self) -> 'FullDictDataClass':
        """复制实例，同 dict.copy()"""
        return self.__class__(self.to_dict())
    
    def setdefault(self, key: str, default: Any = None) -> Any:
        """若键不存在则设置默认值，同 dict.setdefault()"""
        return self.__dict__.setdefault(key, default)
    
    # ----------------------
    # 辅助方法
    # ----------------------
    def to_dict(self) -> Dict[str, Any]:
        """转换为纯字典"""
        return self.__dict__.copy()
    
    def __len__(self) -> int:
        """支持 len(obj) 获取键值对数量"""
        return len(self.__dict__)
    
    def __contains__(self, key: str) -> bool:
        """支持 'key' in obj 判断键是否存在"""
        return key in self.__dict__
    
    def __repr__(self) -> str:
        """实例打印格式"""
        return f"{self.__class__.__name__}({self.__dict__})"
    
    def __str__(self) -> str:
        """实例打印格式"""
        return f"{self.__dict__}"
    
# ----------------------
# 具体数据类（继承增强版基类）
# ----------------------
class Account(FullDictDataClass):
    """账户数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #经纪公司代码
        self.BrokerID = ""
        #投资者帐号
        self.AccountID = ""
        #上次质押金额
        self.PreMortgage = 0
        #上次信用额度
        self.PreCredit = 0
        #上次存款额
        self.PreDeposit = 0
        #上次结算准备金
        self.PreBalance = 0
        #上次占用的保证金
        self.PreMargin = 0
        #利息基数
        self.InterestBase = 0
        #利息收入
        self.Interest = 0
        #入金金额
        self.Deposit = 0
        #出金金额
        self.Withdraw = 0
        #冻结的保证金
        self.FrozenMargin = 0
        #冻结的资金
        self.FrozenCash = 0
        #冻结的手续费
        self.FrozenCommission = 0
        #当前保证金总额
        self.CurrMargin = 0
        #资金差额
        self.CashIn = 0
        #手续费
        self.Commission = 0
        #平仓盈亏
        self.CloseProfit = 0
        #持仓盈亏
        self.PositionProfit = 0
        #期货结算准备金
        self.Balance = 0
        #可用资金
        self.Available = 0
        #可取资金
        self.WithdrawQuota = 0
        #基本准备金
        self.Reserve = 0
        #交易日
        self.TradingDay = 0
        #结算编号
        self.SettlementID = ""
        #信用额度
        self.Credit = 0
        #质押金额
        self.Mortgage = 0
        #交易所保证金
        self.ExchangeMargin = 0
        #投资者交割保证金
        self.DeliveryMargin = 0
        #交易所交割保证金
        self.ExchangeDeliveryMargin = 0
        #保底期货结算准备金
        self.ReserveBalance = 0
        #币种代码
        self.CurrencyID = "CNY"
        #上次货币质入金额
        self.PreFundMortgageIn = 0
        #上次货币质出金额
        self.PreFundMortgageOut = 0
        #货币质入金额
        self.FundMortgageIn = 0
        #货币质出金额
        self.FundMortgageOut = 0
        #货币质押余额
        self.FundMortgageAvailable = 0
        #可质押货币金额--已废弃
        self.MortgageableFund = 0
        #特殊产品占用保证金--已废弃
        self.SpecProductMargin = 0
        #特殊产品冻结保证金--已废弃 
        self.SpecProductFrozenMargin = 0
        #特殊产品手续费--已废弃
        self.SpecProductCommission = 0
        #特殊产品冻结手续费--已废弃
        self.SpecProductFrozenCommission = 0
        #特殊产品持仓盈亏--已废弃
        self.SpecProductPositionProfit = 0
        #特殊产品平仓盈亏--已废弃
        self.SpecProductCloseProfit = 0
        #根据持仓盈亏算法计算的特殊产品持仓盈亏--已废弃
        self.SpecProductPositionProfitByAlg = 0
        #特殊产品交易所保证金--已废弃
        self.SpecProductExchangeMargin = 0
        #业务类型
        self.BizType = ""
        #延时换汇冻结金额
        self.FrozenSwap = 0
        #剩余换汇额度
        self.RemainSwap = 0
        #期权市值
        self.OptionValue = 0
        #本地时间戳
        self.local_timestamp = 0
        #风险度
        self.risk_ratio = 0
        #浮动盈亏
        self.float_profit = 0

class Position(FullDictDataClass):
    """持仓数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #多头持仓
        self.pos_long = 0
        #多头今仓
        self.pos_long_today = 0
        #多头昨仓
        self.pos_long_his = 0
        #多头开仓均价
        self.open_price_long = float("nan")
        #多头持仓均价
        self.position_price_long = float("nan")
        #空头持仓
        self.pos_short = 0
        #空头今仓
        self.pos_short_today = 0
        #空头昨仓
        self.pos_short_his = 0
        #空头开仓均价
        self.open_price_short = float("nan")
        #空头持仓均价
        self.position_price_short = float("nan")
        #多头持仓盈亏
        self.position_profit_long = float("nan")
        #多头浮动盈亏
        self.float_profit_long = float("nan")
        #空头持仓盈亏
        self.position_profit_short = float("nan")
        #空头浮动盈亏
        self.float_profit_short = float("nan")
        #多空总保证金
        self.margin = float("nan")
        #交易所多空总保证金
        self.exch_margin = float("nan")
        #多头保证金
        self.margin_long = float("nan")
        #空头保证金
        self.margin_short = float("nan")
        #多头保证金率(按金额)
        self.margin_rate_long = float("nan")
        #空头保证金率(按金额)
        self.margin_rate_short = float("nan")
        #多头保证金率(按手数)
        self.margin_volume_long = float("nan")
        #空头保证金率(按手数)
        self.margin_volume_short = float("nan")
        #合约代码
        self.instrument_id = ""
        #交易所代码
        self.exchange_id = ""
        #期货"FUTURE",看涨期权"CALL",看跌期权"PUT"
        self.ins_class = ""
        #期权行权价
        self.strike_price  = float("nan")
        #空头今仓冻结
        self.short_frozen_today = 0
        #多头今仓冻结
        self.long_frozen_today = 0
        #空头昨仓冻结
        self.short_frozen_his = 0
        #多头昨仓冻结
        self.long_frozen_his = 0
        #开仓手续费率(按金额)
        self.OpenRatioByMoney = 0
        #开仓手续费(按手数)
        self.OpenRatioByVolume = 0
        #平仓手续费率(按金额)
        self.CloseRatioByMoney = 0
        #平仓手续费(按手数)
        self.CloseRatioByVolume = 0
        #平今仓手续费率(按金额)
        self.CloseTodayRatioByMoney = 0
         #平今仓手续费(按手数)
        self.CloseTodayRatioByVolume = 0
        #本地时间戳
        self.local_timestamp = 0

class Trade(FullDictDataClass):
    """成交单数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #经纪公司代码
        self.BrokerID = ""
        #投资者代码
        self.InvestorID = ""
        #保留的无效字段
        self.reserve1 = ""
        #报单引用
        self.OrderRef = ""
        #用户代码
        self.UserID = ""
        #交易所代码
        self.ExchangeID = ""
        #成交编号
        self.TradeID = ""
        #买卖方向
        self.Direction = ""
        #报单编号
        self.OrderSysID = ""
        #会员代码
        self.ParticipantID = ""
        #客户代码
        self.ClientID = ""
        #交易角色
        self.TradingRole = ""
        #保留的无效字段
        self.reserve2 = ""
        #开平标志
        self.OffsetFlag = ""
        #投机套保标志
        self.HedgeFlag = ""
        #价格
        self.Price = float("nan")
        #数量
        self.Volume = float("nan")
        #成交时期
        self.TradeDate = ""
        #成交时间
        self.TradeTime = ""
        #成交类型
        self.TradeType = ""
        #成交价来源
        self.PriceSource = ""
        #交易所交易员代码
        self.TraderID = ""
        #本地报单编号
        self.OrderLocalID = ""
        #结算会员编号
        self.ClearingPartID = ""
        #业务单元
        self.BusinessUnit = ""
        #序号
        self.SequenceNo = ""
        #交易日
        self.TradingDay = ""
        #结算编号
        self.SettlementID = ""
        #经纪公司报单编号
        self.BrokerOrderSeq = ""
        #成交来源
        self.TradeSource = ""
        #投资单元代码
        self.InvestUnitID = ""
        #合约代码
        self.InstrumentID = ""
        #合约在交易所的代码
        self.ExchangeInstID = ""
        #本地时间戳
        self.local_timestamp = 0

class Order(FullDictDataClass):
    """委托单数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #经纪公司代码
        self.BrokerID = ""
        #投资者代码
        self.InvestorID = ""
        #保留的无效字段
        self.reserve1 = ""
        #报单引用
        self.OrderRef = ""
        #用户代码
        self.UserID = ""
        #报单价格条件
        self.OrderPriceType = ""
        #买卖方向
        self.Direction = ""
        #组合开平标志
        self.CombOffsetFlag = ""
        #组合投机套保标志
        self.CombHedgeFlag = ""
        #价格
        self.LimitPrice = float("nan")
        #数量
        self.VolumeTotalOriginal = float("nan")
        #有效期类型
        self.TimeCondition = ""
        #GTD日期
        self.GTDDate = ""
        #成交量类型
        self.VolumeCondition = ""
        #最小成交量
        self.MinVolume = float("nan")
        #触发条件
        self.ContingentCondition = ""
        #止损价
        self.StopPrice = ""
        #强平原因
        self.ForceCloseReason = ""
        #自动挂起标志
        self.IsAutoSuspend = ""
        #业务单元
        self.BusinessUnit = ""
        #请求编号
        self.RequestID = ""
        #本地报单编号
        self.OrderLocalID = ""
        #交易所代码
        self.ExchangeID = ""
        #会员代码
        self.ParticipantID = ""
        #客户代码
        self.ClientID = ""
        #保留的无效字段
        self.reserve2 = ""
        #交易所交易员代码
        self.TraderID = ""
        #安装编号
        self.InstallID = ""
        #报单提交状态
        self.OrderSubmitStatus = ""
        #报单提示序号
        self.NotifySequence = ""
        #交易日
        self.TradingDay = ""
        #结算编号
        self.SettlementID = ""
        #报单编号
        self.OrderSysID = ""
        #报单来源
        self.OrderSource = ""
        #报单状态
        self.OrderStatus = ""
        #报单类型
        self.OrderType = ""
        #今成交数量
        self.VolumeTraded = float("nan")
        #剩余数量
        self.VolumeTotal = float("nan")
        #报单日期
        self.InsertDate = ""
        #委托时间
        self.InsertTime = ""
        #激活时间
        self.ActiveTime = ""
        #挂起时间
        self.SuspendTime = ""
        #最后修改时间
        self.UpdateTime = ""
        #撤销时间
        self.CancelTime = ""
        #最后修改交易所交易员代码
        self.ActiveTraderID = ""
        #结算会员编号
        self.ClearingPartID = ""
        #序号
        self.SequenceNo = ""
        #前置编号
        self.FrontID = ""
        #会话编号
        self.SessionID = ""
        #用户端产品信息
        self.UserProductInfo = ""
        #状态信息
        self.StatusMsg = ""
        #用户强平标志
        self.UserForceClose = ""
        #操作用户代码
        self.ActiveUserID = ""
        #经纪公司报单编号
        self.BrokerOrderSeq = ""
        #相关报单
        self.RelativeOrderSysID = ""
        #郑商所成交数量
        self.ZCETotalTradedVolume = ""
        #互换单标志
        self.IsSwapOrder = ""
        #营业部编号
        self.BranchID = ""
        #投资单元代码
        self.InvestUnitID = ""
        #资金账号
        self.AccountID = ""
        #币种代码
        self.CurrencyID = ""
        #保留的无效字段
        self.reserve3 = ""
        #Mac地址
        self.MacAddress = ""
        #合约代码
        self.InstrumentID = ""
        #合约在交易所的代码
        self.ExchangeInstID = ""
        #IP地址
        self.IPAddress = ""
        #报单回显字段
        self.OrderMemo = ""
        #session上请求计数 api自动维护
        self.SessionReqSeq = ""
        #本地时间戳
        self.local_timestamp = 0

class Quote(FullDictDataClass):
    """行情快照数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #交易日
        self.TradingDay = ""
        #合约代码
        self.InstrumentID = ""
        #交易所代码
        self.ExchangeID = ""
        #合约在交易所的代码
        self.ExchangeInstID = ""
        #最新价
        self.LastPrice = float("nan")
        #上次结算价
        self.PreSettlementPrice = float("nan")
        #昨收盘
        self.PreClosePrice = float("nan")
        #昨持仓量
        self.PreOpenInterest = float("nan")
        #今开盘
        self.OpenPrice = float("nan")
        #最高价
        self.HighestPrice = float("nan")
        #最低价
        self.LowestPrice = float("nan")
        #数量
        self.Volume = float("nan")
        #成交金额
        self.Turnover = float("nan")
        #持仓量
        self.OpenInterest = float("nan")
        #今收盘
        self.ClosePrice = float("nan")
        #本次结算价
        self.SettlementPrice = float("nan")
        #涨停板价
        self.UpperLimitPrice = float("nan")
        #跌停板价
        self.LowerLimitPrice = float("nan")
        #昨虚实度
        self.PreDelta = float("nan")
        #今虚实度
        self.CurrDelta = float("nan")
        #最后修改时间
        self.UpdateTime = ""
        #最后修改毫秒
        self.UpdateMillisec = float("nan")
        #申买价一
        self.BidPrice1 = float("nan")
        #申买量一
        self.BidVolume1 = float("nan")
        #申卖价一
        self.AskPrice1 = float("nan")
        #申卖量一
        self.AskVolume1 = float("nan")
        #申买价二
        self.BidPrice2 = float("nan")
        #申买量二
        self.BidVolume2 = float("nan")
        #申卖价二
        self.AskPrice2 = float("nan")
        #申卖量二
        self.AskVolume2 = float("nan")
        #申买价三
        self.BidPrice3 = float("nan")
        #申买量三
        self.BidVolume3 = float("nan")
        #申卖价三
        self.AskPrice3 = float("nan")
        #申卖量三
        self.AskVolume3 = float("nan")
        #申买价四
        self.BidPrice4 = float("nan")
        #申买量四
        self.BidVolume4 = float("nan")
        #申卖价四
        self.AskPrice4 = float("nan")
        #申卖量四
        self.AskVolume4 = float("nan")
        #申买价五
        self.BidPrice5 = float("nan")
        #申买量五
        self.BidVolume5 = float("nan")
        #申卖价五
        self.AskPrice5 = float("nan")
        #申卖量五
        self.AskVolume5 = float("nan")
        #当日均价
        self.AveragePrice = float("nan")
        #业务日期
        self.ActionDay = float("nan")
        #上带价
        self.BandingUpperPrice = float("nan")
        #下带价
        self.BandingLowerPrice = float("nan")
        #本地时间戳
        self.local_timestamp = 0
        #CTP行情时间戳
        self.ctp_timestamp = 0
        #CTP行情datetime
        self.ctp_datetime = 0

class InstrumentProperty(FullDictDataClass):
    """行情快照数据类"""
    def __init__( self,
        # 允许动态传入其他字段（**kwargs）
        **kwargs
    ):
        #  处理动态扩展字段（通过基类初始化）
        super().__init__(kwargs)  # 等价于：self.__dict__.update(kwargs)
        #保留的无效字段
        self.reserve1 = ""
        #交易所代码
        self.ExchangeID = ""
        #合约名称
        self.InstrumentName = ""
        #保留的无效字段
        self.reserve2 = ""
        #保留的无效字段
        self.reserve3 = ""
        #产品类型
        self.ProductClass = ""
        #交割年份
        self.DeliveryYear = ""
        #交割月
        self.DeliveryMonth = ""
        #市价单最大下单量
        self.MaxMarketOrderVolume = float("nan")
        #市价单最小下单量
        self.MinMarketOrderVolume = float("nan")
        #限价单最大下单量
        self.MaxLimitOrderVolume = float("nan")
        #限价单最小下单量
        self.MinLimitOrderVolume = float("nan")
        #合约数量乘数
        self.VolumeMultiple = float("nan")
        #最小变动价位
        self.PriceTick = float("nan")
        #创建日
        self.CreateDate = ""
        #上市日
        self.OpenDate = ""
        #到期日
        self.ExpireDate = ""
        #开始交割日
        self.StartDelivDate = ""
        #结束交割日
        self.EndDelivDate = ""
        #合约生命周期状态
        self.InstLifePhase = ""
        #当前是否交易
        self.IsTrading = ""
        #持仓类型
        self.PositionType = ""
        #持仓日期类型
        self.PositionDateType = float("nan")
        #多头保证金率
        self.LongMarginRatio = 0
        #空头保证金率
        self.ShortMarginRatio = 0
        #是否使用大额单边保证金算法
        self.MaxMarginSideAlgorithm = ""
        #保留的无效字段
        self.reserve4 = ""
        #执行价
        self.StrikePrice = float("nan")
        #期权类型
        self.OptionsType = ""
        #合约基础商品乘数
        self.UnderlyingMultiple = float("nan")
        #组合类型
        self.CombinationType = ""
        #合约代码
        self.InstrumentID = ""
        #合约在交易所的代码
        self.ExchangeInstID = ""
        #产品代码
        self.ProductID = ""
        #基础商品代码
        self.UnderlyingInstrID = ""
        #到期剩余日
        self.expire_rest_days = float('nan')



# 配置参数
# 支持的时间单位（与Polars兼容）
SUPPORTED_UNITS = {
    "s": {"full_name": "seconds", "desc": "秒"},    # 周期缩写→完整单位名称映射
    "m": {"full_name": "minutes", "desc": "分钟"},
    "h": {"full_name": "hours", "desc": "小时"},
    "d": {"full_name": "days", "desc": "天"},
    "w": {"full_name": "weeks", "desc": "周"},
    "M": {"full_name": "months", "desc": "月"},     # 注意：pl.duration不支持months，需特殊处理
    "y": {"full_name": "years", "desc": "年"}       # pl.duration不支持years，需特殊处理
}
# 预定义常用周期（可扩展）
COMMON_CYCLES = ["1s", "3s", "5s", "10s", "30s", "1m", "5m", "15m", "30m", "1h", "2h", "4h", "1d", "1w"]
# 区分日内/日级周期的单位阈值
INTRADAY_UNITS = {"s", "m", "h"}  # 日内周期单位
DAILY_UNITS = {"d", "w", "M", "y"}  # 日及以上周期单位

class MutableDataHolder():
    """通用可变数据容器基类，封装DataFrame引用实现自动更新"""
    def __init__(self, schema: Dict[str, polars.DataType]):
        # 初始化空DataFrame（使用指定schema确保结构一致性）
        self._data: polars.DataFrame = polars.DataFrame(
            {col: [] for col in schema.keys()},
            schema=schema
        )

    @property
    def data(self) -> polars.DataFrame:
        """获取最新数据（返回引用，不复制）"""
        return self._data

    @data.setter
    def data(self, new_data: polars.DataFrame) -> None:
        """更新数据（替换引用，不修改原对象）"""
        # 校验列结构一致性
        required_cols = set(self._data.columns)
        if not required_cols.issubset(set(new_data.columns)):
            missing_cols = required_cols - set(new_data.columns)
            raise ValueError(f"新数据缺少必要列：{missing_cols}")
        
        self._data = new_data

    def tail(self, n: int) -> polars.DataFrame:
        """便捷方法：获取最新n条数据"""
        return self._data.tail(n)


class MutableTickHolder(MutableDataHolder):
    """Tick数据可变容器，定义Tick专属schema"""
    def __init__(self):
        tick_schema = {
            "TradingDay": polars.String,
            "ActionDay": polars.String,
            "InstrumentID": polars.String,
            "ExchangeID": polars.String,
            "LastPrice": polars.Float64,
            "Volume": polars.Int64,
            "Turnover": polars.Float64,
            "OpenInterest": polars.Float64,
            "BidPrice1": polars.Float64,
            "BidVolume1": polars.Int64,
            "AskPrice1": polars.Float64,
            "AskVolume1": polars.Int64,
            "UpdateTime": polars.String,
            "UpdateMillisec": polars.Int64,
            "ctp_datetime": polars.Datetime,
            "trading_day": polars.Date,
            "is_valid": polars.Boolean
        }
        super().__init__(tick_schema)


class MutableKlineHolder(MutableDataHolder):
    """K线数据可变容器，定义K线专属schema"""
    def __init__(self):
        kline_schema = {
            "InstrumentID": polars.String,
            "open": polars.Float64,
            "high": polars.Float64,
            "low": polars.Float64,
            "close": polars.Float64,
            "Volume": polars.Int64,
            "Turnover": polars.Float64,
            "OpenInterest": polars.Float64,
            "period_start": polars.Datetime,
            "period_end": polars.Datetime,
            "period": polars.String
        }
        super().__init__(kline_schema)

class CTPDataProcessor():
    def __init__(self, storage_format: str = "parquet",_logfile=""):
        """初始化处理器，支持选择存储格式（parquet或csv）"""
        if not _logfile:
            try:
                self._flowfile = os.path.dirname(os.path.abspath(__file__)) #当前程序目录(该py文件__file__所在目录)
            except: self._flowfile = os.getcwd()   #程序工作目录下创建目录 
            self._logfile = fr"{self._flowfile}\logs"
        else: self._logfile = _logfile
        self.tick_dir = fr"{self._logfile}\tick_data"
        self.kline_dir = fr"{self._logfile}\kline_data"
        # 创建数据目录
        os.makedirs(self.tick_dir, exist_ok=True)
        os.makedirs(self.kline_dir, exist_ok=True)
        
        # 验证存储格式
        self.storage_format = storage_format.lower()
        if self.storage_format not in ["parquet", "csv"]:
            raise ValueError(f"不支持的存储格式：{storage_format}，仅支持 'parquet' 和 'csv'")
        
        # 数据缓存：使用可变容器
        self.tick_cache: Dict[str, MutableTickHolder] = {}  # {合约ID: 可变Tick容器}
        self.kline_cache: Dict[str, Dict[str, MutableKlineHolder]] = {}  # {合约ID: {周期: 可变K线容器}}
        self.kline_periods = {}  #K线周期列表
        self.max_tick_count = {}  #最大缓存Tick数量
        self.max_kline_count = {}  #最大缓存kline数量

        # 加载本地数据
        self._load_local_data()

    # 工具方法：周期单位转换 
    def _get_duration_param(self, unit: str, value: int) -> Dict[str, int]:
        """
        将周期缩写（如'm'）转换为pl.duration()的合法参数
        返回格式：{完整单位名称: 数值}，如{"minutes": 5}
        """
        if unit not in SUPPORTED_UNITS:
            raise ValueError(f"不支持的周期单位：{unit}")
        
        full_unit = SUPPORTED_UNITS[unit]["full_name"]
        # 特殊处理：pl.duration不支持months/years，用days近似（或根据业务调整）
        if full_unit in ["months", "years"]:
            raise ValueError(f"pl.duration不支持{SUPPORTED_UNITS[unit]['desc']}单位，请使用dt.offset_by替代")
        
        return {full_unit: value}
    # 文件路径管理
    def _get_tick_file_path(self, instrument_id: str) -> str:
        """获取Tick文件路径（根据存储格式自动选择扩展名）"""
        return os.path.join(self.tick_dir, f"{instrument_id}_tick.{self.storage_format}")

    def _get_kline_file_path(self, instrument_id: str, period: str) -> str:
        """获取K线文件路径（根据存储格式自动选择扩展名）"""
        period_dir = os.path.join(self.kline_dir, period)
        os.makedirs(period_dir, exist_ok=True)
        return os.path.join(period_dir, f"{instrument_id}_kline.{self.storage_format}")

    # 本地数据加载
    def _load_local_data(self) -> None:
        """加载本地所有合约的历史数据"""
        # 查找对应格式的Tick文件
        tick_files = glob.glob(os.path.join(self.tick_dir, f"*.{self.storage_format}"))
        
        for file_path in tick_files:
            try:
                # 解析合约ID（文件名格式：{合约ID}_tick.xxx）
                file_name = os.path.basename(file_path)
                instrument_id = file_name.replace(f"_tick.{self.storage_format}", "")
                
                # 读取并解析Tick数据
                tick_df = self._read_tick_file(file_path)
                #print(f"加载 {instrument_id} 历史Tick数据：共 {len(tick_df)} 条")

                # 初始化Tick容器
                tick_holder = MutableTickHolder()
                tick_holder.data = tick_df
                self.tick_cache[instrument_id] = tick_holder
                
            except Exception as e:
                print(f"加载 {file_path} 失败：{str(e)}")

    def _read_tick_file(self, file_path: str) -> polars.DataFrame:
        """读取本地Tick文件（自动适配parquet和csv格式）"""
        try:
            if file_path.endswith(".parquet"):
                # 读取Parquet文件（保留原始数据类型）
                tick_df = polars.read_parquet(file_path)
            else:
                # 读取CSV文件（需手动解析时间字段）
                tick_df = polars.read_csv(
                    file_path,
                    parse_dates=["ctp_datetime", "trading_day"]
                )
            
            # 确保时间类型正确
            if "ctp_datetime" in tick_df.columns and not isinstance(tick_df["ctp_datetime"].dtype, polars.Datetime):
                tick_df = tick_df.with_columns(
                    polars.col("ctp_datetime").str.strptime(polars.Datetime, "%Y-%m-%d %H:%M:%S%.f").alias("ctp_datetime")
                )
            if "trading_day" in tick_df.columns and not isinstance(tick_df["trading_day"].dtype, polars.Date):
                tick_df = tick_df.with_columns(
                    polars.col("trading_day").str.strptime(polars.Date, "%Y-%m-%d").alias("trading_day")
                )
                
            return tick_df.sort("ctp_datetime")
            
        except Exception as e:
            raise ValueError(f"解析Tick文件 {file_path} 失败：{str(e)}")

    # Tick数据处理
    def process_snapshot(self, snapshot, save_instrument_id: str = "") -> None:
        """处理单条CTP快照数据"""
        
        instrument_id = snapshot['InstrumentID']
        if not save_instrument_id: save_instrument_id = instrument_id
        tick_record = self._snapshot_to_tick(snapshot)
        self._update_tick_cache(instrument_id, tick_record)
        self._append_tick_to_file(save_instrument_id, tick_record)
        # 更新所有周期K线
        for period in self.kline_periods[instrument_id]:
            self._update_kline(instrument_id, period,self.max_kline_count[instrument_id][period])
        #print({"process_snapshot":snapshot})

    def process_snapshots_batch(self, snapshots: List, save_instrument_id: str = "") -> None:
        """批量处理CTP快照（提升高频场景性能）"""
        if not snapshots:
            return
            
        # 按合约分组处理
        from collections import defaultdict
        grouped_snapshots = defaultdict(list)
        for snap in snapshots:
            grouped_snapshots[snap['InstrumentID']].append(snap)
        
        for instrument_id, snaps in grouped_snapshots.items():
            # 批量转换为Tick
            tick_records = [self._snapshot_to_tick(snap) for snap in snaps]
            new_ticks = polars.DataFrame(tick_records)
            
            # 批量更新缓存和文件
            if not save_instrument_id: save_instrument_id = instrument_id
            self._update_tick_cache_batch(instrument_id, new_ticks)
            self._append_ticks_batch(save_instrument_id, new_ticks)
            
            # 更新K线
            for period in self.kline_periods[instrument_id]:
                self._update_kline(instrument_id, period,self.max_kline_count[instrument_id][period])

    def _snapshot_to_tick(self, snapshot) -> Dict:
        """将CTP快照转换为标准化Tick（含双重时间标识）"""
        # 1. 自然时间（物理时间流）
        action_day = snapshot["ActionDay"]  # 格式："20231010"
        try:
            natural_time_str = f"{action_day} {snapshot['UpdateTime']}.{snapshot['UpdateMillisec']:03d}"
            ctp_datetime = datetime.strptime(natural_time_str, "%Y%m%d %H:%M:%S.%f")
        except ValueError as e:
            ctp_datetime = datetime.strptime(f"{action_day} {snapshot['UpdateTime']}", "%Y%m%d %H:%M:%S")
        
        # 2. 交易日标识（业务日期）
        trading_day = snapshot['TradingDay']  # 格式："20231011"
        trading_day_date = datetime.strptime(trading_day, "%Y%m%d").date()
        snapshot.update({"ctp_datetime": ctp_datetime,"trading_day": trading_day_date, "is_valid": snapshot['LastPrice'] > 0})
        return snapshot

    def _update_tick_cache(self, instrument_id: str, tick_record: Dict) -> None:
        """更新单个Tick到缓存"""
        new_tick = polars.DataFrame([tick_record])
        self._update_tick_cache_batch(instrument_id, new_tick)
        

    def _update_tick_cache_batch(self, instrument_id: str, new_ticks: polars.DataFrame) -> None:
        """批量更新Tick缓存"""
        # 初始化容器（首次更新时）
        if instrument_id not in self.tick_cache:
            self.tick_cache[instrument_id] = MutableTickHolder()
        
        tick_holder = self.tick_cache[instrument_id]
        current_ticks = tick_holder.data
        # 合并数据并去重
        if len(current_ticks) > 0:
            combined = polars.concat([current_ticks, new_ticks])
            combined = combined.sort("ctp_datetime").unique(subset=["ctp_datetime"], keep="last")
            # 限制缓存大小
            if len(combined) > self.max_tick_count[instrument_id]:
                combined = combined.tail(self.max_tick_count[instrument_id])
            tick_holder.data = combined
        else:
            # 首次添加
            tick_holder.data = new_ticks.sort("ctp_datetime")

    # Tick数据持久化（双格式支持）
    def _append_tick_to_file(self, instrument_id: str, tick_record: Dict) -> None:
        """追加单个Tick到文件"""
        self._append_ticks_batch(instrument_id, polars.DataFrame([tick_record]))

    def _append_ticks_batch(self, instrument_id: str, tick_df: polars.DataFrame) -> None:
        """批量追加Tick到文件（根据存储格式选择方式）"""
        file_path = self._get_tick_file_path(instrument_id)
        
        # 格式化时间字段（便于存储）
        formatted_df = tick_df.with_columns([
            polars.col("ctp_datetime").dt.strftime("%Y-%m-%d %H:%M:%S%.f").alias("ctp_datetime"),
            polars.col("trading_day").dt.strftime("%Y-%m-%d").alias("trading_day")
        ])
        
        if self.storage_format == "parquet":
            # Parquet不支持追加，需先读再合并（适合大数据量）
            if os.path.exists(file_path):
                existing_df = polars.read_parquet(file_path)
                combined_df = polars.concat([existing_df, formatted_df])
                # 去重并限制大小
                combined_df = combined_df.unique(subset=["ctp_datetime"], keep="last").sort("ctp_datetime")
                #if len(combined_df) > self.max_tick_count[instrument_id]:
                #    combined_df = combined_df.tail(self.max_tick_count[instrument_id])
                combined_df.write_parquet(file_path)
            else:
                formatted_df.write_parquet(file_path)
        else:
            # CSV支持追加（适合小数据量或调试）
            include_header = not os.path.exists(file_path)
            with open(file_path, "a", encoding="utf-8") as f:
                formatted_df.write_csv( f, include_header=include_header )

    # K线生成与更新
    @staticmethod
    def _validate_period(period: str) -> Tuple[bool, Optional[str]]:
        """验证周期格式合法性"""
        pattern = re.compile(r"^(\d+)([smhdwMy])$")
        match = pattern.match(period)
        if not match:
            return False, f"周期格式错误：{period}（正确格式：数字+单位，如'1s'）"
        
        unit = match.group(2)
        if unit not in SUPPORTED_UNITS:
            return False, f"不支持的单位：{unit}（支持：{list(SUPPORTED_UNITS.keys())}）"
        
        return True, None

    @staticmethod
    def _classify_period(period: str) -> Tuple[str, int, str]:
        """解析周期为（类型, 数值, 单位）"""
        pattern = re.compile(r"^(\d+)([smhdwMy])$")
        match = pattern.match(period)
        if not match:
            raise ValueError(f"无效周期格式：{period}")
        
        num = int(match.group(1))
        unit = match.group(2)
        period_type = "intraday" if unit in INTRADAY_UNITS else "daily"
        
        return period_type, num, unit

    def generate_any_period_kline( self, instrument_id: str, tick_df: polars.DataFrame, period: str, kline_count: int ) -> polars.DataFrame:
        """生成任意周期K线（日内用自然时间，日级用交易日）"""
        # 验证周期
        is_valid, msg = self._validate_period(period)
        if not is_valid:
            raise ValueError(f"生成K线失败：{msg}")
        
        # 解析周期类型
        period_type, num, unit = self._classify_period(period)

        # 选择时间基准
        if period_type == "intraday":
            time_col = "ctp_datetime"
            if not isinstance(tick_df[time_col].dtype, polars.Datetime):
                tick_df = tick_df.with_columns(polars.col(time_col).cast(polars.Datetime).alias(time_col))
        else:
            time_col = "trading_day"
            tick_df = tick_df.with_columns(polars.col(time_col).cast(polars.Date).alias(time_col))

        # 过滤必要Tick
        filtered_ticks = self._filter_needed_ticks(tick_df, time_col, period, self.max_kline_count[instrument_id][period], instrument_id)
        self.max_tick_count[instrument_id] = max(self.max_tick_count[instrument_id], len(filtered_ticks))
        # 生成K线
        if period_type == "intraday":
            kline_df = (
                filtered_ticks.group_by_dynamic(
                    time_col, by="InstrumentID", every=period, closed="left", include_boundaries=True
                )
                .agg(self._get_kline_aggs(time_col))
                .with_columns(polars.lit(period).alias("period"))
                .sort(time_col)
            )
        else:
            kline_df = (
                filtered_ticks.group_by([polars.col("InstrumentID"), polars.col(time_col).alias("kline_date")])
                .agg(self._get_kline_aggs(time_col))
                .with_columns([
                    polars.lit(period).alias("period"),
                    polars.col("kline_date").alias("kline_start_time"),
                    polars.col("kline_date").alias("kline_end_time")
                ])
                .sort("kline_date")
            )
        
        # 限制数量
        return kline_df.tail(kline_count) if len(kline_df) > kline_count else kline_df

    def _get_kline_aggs(self, time_col: str) -> List:
        """K线聚合逻辑"""
        return [
            polars.col("LastPrice").first().alias("open"),
            polars.col("LastPrice").max().alias("high"),
            polars.col("LastPrice").min().alias("low"),
            polars.col("LastPrice").last().alias("close"),
            (polars.col("Volume").last() - polars.col("Volume").first()).alias("Volume"),
            (polars.col("Turnover").last() - polars.col("Turnover").first()).alias("Turnover"),
            polars.col("OpenInterest").last().alias("OpenInterest"),
            polars.col(time_col).first().alias("period_start"),
            polars.col(time_col).last().alias("period_end")
        ]

    def _filter_needed_ticks(self, tick_df: polars.DataFrame, time_col: str, period: str, kline_count: int,instrument_id="") -> polars.DataFrame:
        """过滤生成K线所需的Tick数据"""
        if len(tick_df) == 0:
            return tick_df
        
        period_type, num, unit = self._classify_period(period)
        buffer_periods = 2
        total_offset_value = num * (kline_count + buffer_periods)

        latest_time = tick_df.select(polars.col(time_col).max()).item()
        tick_earliest_time = tick_df.select(polars.col(time_col).min()).item()

        # 计算最早保留时间（加缓冲避免边界丢失）
        if period_type == "intraday":
            # 日内周期：用pl.duration计算固定时间偏移
            duration_param = self._get_duration_param(unit, total_offset_value)
            earliest_expr = polars.lit(latest_time) - polars.duration(** duration_param)  # 传入合法参数
            earliest_time = tick_df.select(earliest_expr).item()
        else:
            days_per_unit = {"d": 1, "w": 7, "M": 30, "y": 365}.get(unit, 1)
            earliest_time = latest_time - timedelta(days=total_offset_value * days_per_unit)
        if tick_earliest_time > earliest_time: #数据量不足重新读取本地数据
            #instrument_id = tick_df.tail(1).get_column("InstrumentID").item()
            self._load_instrument_data(instrument_id)
            tick_df = self.tick_cache[instrument_id].data
            
        return tick_df.filter(polars.col(time_col) >= earliest_time)

    def _update_kline(self, instrument_id: str, period: str, kline_count: int  ) -> None:
        """更新K线缓存"""
        if instrument_id not in self.tick_cache:
            return

        # 初始化K线容器
        if instrument_id not in self.kline_cache:
            self.kline_cache[instrument_id] = {}
        if period not in self.kline_cache[instrument_id]:
            self.kline_cache[instrument_id][period] = MutableKlineHolder()

        # 获取数据
        tick_holder = self.tick_cache[instrument_id]
        tick_df = tick_holder.data
        kline_holder = self.kline_cache[instrument_id][period]
        current_kline = kline_holder.data

        # 首次生成
        if len(current_kline) == 0:
            new_kline = self.generate_any_period_kline(instrument_id, tick_df, period, kline_count)
            kline_holder.data = new_kline
            self._save_kline_to_file(instrument_id, period, new_kline)
            return

        # 增量更新
        period_type, _, _ = self._classify_period(period)
        time_col = "ctp_datetime" if period_type == "intraday" else "trading_day"
        latest_period_end = current_kline.select(polars.col("period_end").max()).item()
        new_ticks = tick_df.filter(polars.col(time_col) > latest_period_end)

        if len(new_ticks) == 0:
            return

        new_klines = self.generate_any_period_kline(instrument_id, new_ticks, period, kline_count + 5)
        if len(new_klines) == 0:
            return
        
        # 合并并去重
        combined_kline = current_kline.vstack(new_klines).unique(subset=["period_start"], keep="last")
        combined_kline = combined_kline.sort("period_start")
        if len(combined_kline) > kline_count:
            combined_kline = combined_kline.tail(kline_count)

        # 更新容器
        kline_holder.data = combined_kline
        self._save_kline_to_file(instrument_id, period, combined_kline)

    # K线持久化（双格式支持）
    def _save_kline_to_file(self, instrument_id: str, period: str, kline_df: polars.DataFrame) -> None:
        return
        """保存K线到文件"""
        file_path = self._get_kline_file_path(instrument_id, period)

        # 格式化时间字段
        formatted_df = kline_df.with_columns([
            polars.col("period_start").cast(polars.String).alias("period_start"),
            polars.col("period_end").cast(polars.String).alias("period_end"),
            polars.col("kline_start_time").cast(polars.String).alias("kline_start_time") 
            if "kline_start_time" in kline_df.columns else polars.lit(None),
            polars.col("kline_end_time").cast(polars.String).alias("kline_end_time")
            if "kline_end_time" in kline_df.columns else polars.lit(None)
        ])

        # 按格式保存
        if self.storage_format == "parquet":
            formatted_df.write_parquet(file_path)
        else:
            include_header = not os.path.exists(file_path)
            with open(file_path, "a", encoding="utf-8") as f:
                formatted_df.write_csv( f, include_header=include_header )

    # 数据获取接口
    def get_tick_holder(self, instrument_id: str) -> Optional[MutableTickHolder]:
        """获取Tick可变容器（外部通过容器获取最新数据）"""
        if instrument_id not in self.max_tick_count: self.max_tick_count[instrument_id] = 10000
        if instrument_id not in self.tick_cache:
            self._load_instrument_data(instrument_id)
        if instrument_id not in self.tick_cache: self.tick_cache[instrument_id] = MutableTickHolder()  
        type(self.tick_cache[instrument_id]),id(self.tick_cache[instrument_id]) 
        return self.tick_cache[instrument_id]#.tail(tick_count)

    def get_kline_holder(self, instrument_id: str, period: str, kline_count:int) -> Optional[MutableKlineHolder]:
        """获取K线可变容器"""
        if instrument_id not in self.max_kline_count: self.max_kline_count[instrument_id] = {period:kline_count}
        elif period not in self.max_kline_count[instrument_id]: self.max_kline_count[instrument_id][period] = kline_count
        else: self.max_kline_count[instrument_id][period] = max(self.max_kline_count[instrument_id][period], kline_count)
        if instrument_id not in self.max_tick_count: self.max_tick_count[instrument_id] = 10000
        if instrument_id in self.kline_cache and period in self.kline_cache[instrument_id]:
            return self.kline_cache[instrument_id][period]
        if not self._validate_period(period)[0]:
            print(f"无效周期：{period}")
            return None
        kline_holder = MutableKlineHolder()   #先创建空K线
        self.kline_cache[instrument_id] = {period:kline_holder}
        if instrument_id not in self.kline_periods: #缓存中无周期则添加
            self.kline_periods[instrument_id] = set()
        self.kline_periods[instrument_id].add(period)
        if instrument_id not in self.tick_cache: #缓存中无tick尝试从本地读取
            self._load_instrument_data(instrument_id)
        if instrument_id in self.tick_cache:
            # 生成K线
            kline_data = self.generate_any_period_kline(instrument_id, self.tick_cache[instrument_id].data, period, kline_count)
            kline_holder.data = kline_data
        else:
            tick_holder = MutableTickHolder() #创建空tick
            self.tick_cache[instrument_id] = tick_holder
        #kline =  self.kline_cache.get(instrument_id, {}).get(period)    
        return self.kline_cache[instrument_id][period]

    def _load_instrument_data(self, instrument_id: str) -> None:
        """懒加载指定合约数据"""
        tick_file = self._get_tick_file_path(instrument_id)
        if not os.path.exists(tick_file):
            print(f"未找到 {instrument_id} 的本地文件")
            return

        try:
            # 加载Tick
            tick_df = self._read_tick_file(tick_file)
            if instrument_id in self.tick_cache and isinstance(self.tick_cache[instrument_id],MutableTickHolder): self.tick_cache[instrument_id].data = tick_df
            else:
                tick_holder = MutableTickHolder()
                tick_holder.data = tick_df
                self.tick_cache[instrument_id] = tick_holder
            
            #print(f"懒加载 {instrument_id} 数据：{len(tick_df)} 条Tick")

        except Exception as e:
            print(f"懒加载 {instrument_id} 失败：{str(e)}")

    def get_latest_realtime_tick(self, instrument_id: str) -> Optional[polars.DataFrame]:
        """快捷获取最新Tick"""
        tick_holder = self.get_tick_holder(instrument_id)
        return tick_holder.tail(1) if tick_holder else None





#!/usr/bin/env python
#  -*- coding: utf-8 -*-
import sys
import os
# 获取当前脚本的绝对路径
current_path = os.path.abspath(__file__)
root_path = os.path.abspath(os.path.join(current_path, "../../"))
# 将根目录添加到 sys.path
if root_path not in sys.path:
    sys.path.append(root_path)
from trade_mdforopenctp import PeopleQuantApi
import time as tm
import zhuchannel
import asyncio
import traceback
import types
import polars
from datetime import datetime,time,date,timedelta
import copy
from typing import Dict, List, Optional, Tuple, Any

envs = {
    "7x24": {
        "td": "tcp://182.254.243.31:40001",
        "md": "tcp://182.254.243.31:40011",
    },
    "电信1": {
        "td": "tcp://182.254.243.31:30001",
        "md": "tcp://182.254.243.31:30011",
    },
    "电信2": {
        "td": "tcp://182.254.243.31:30002",
        "md": "tcp://182.254.243.31:30012",
    },
    "移动": {
        "td": "tcp://182.254.243.31:30003",
        "md": "tcp://182.254.243.31:30013",
    },
}
TradeFrontAddr = envs["电信1"]["td"]
MdFrontAddr = envs["电信1"]["md"]
TradeFrontAddr = envs["7x24"]["td"]
MdFrontAddr = envs["7x24"]["md"]

#TradeFrontAddr = "tcp://121.37.80.177:20002" #openctp
#MdFrontAddr = "tcp://121.37.80.177:20004" #openctp

BROKERID="9999"   #期货公司ID
USERID=""   #账户
PASSWORD=""   #登录密码
APPID="simnow_client_test"   #客户端ID
AUTHCODE="0000000000000000"  #授权码

#策略函数
def cta(symbol_ctp ):
    quote1 = pqapi.get_quote(symbol_ctp)        #获取合约行情
    position1 = pqapi.get_position(symbol_ctp)   #获取合约持仓
    task = zhuchannel.WorkThread(pqapi.auto_pos_thread,args=(symbol_ctp,),kwargs={"return_queue":None})
    task.start()
    UpdateTime = quote1.ctp_datetime #行情更新时间
    while True:
        if UpdateTime != quote1.ctp_datetime: #新行情推送
            UpdateTime = quote1.ctp_datetime
            buy_up = 1  #多头信号
            sell_down = 1  #空头信号
            if buy_up :
                pqapi.set_pos_volume(symbol_ctp,pos_long=15)
            elif sell_down :
                pqapi.set_pos_volume(symbol_ctp,pos_short=15)
            if position1.pos_long and abs(quote1.LastPrice - position1.open_price_long) >= 2:
                pqapi.set_pos_volume(symbol_ctp,pos_long=0)
            if position1.pos_short and abs(quote1.LastPrice - position1.open_price_short) >= 2:
                pqapi.set_pos_volume(symbol_ctp,pos_short=0)
                    
#创建api实例
pqapi = PeopleQuantApi()
account = pqapi.get_account()            #获取账户资金
#创建策略线程
cta2 = zhuchannel.WorkThread(cta,args=('hc2601', ),kwargs={})
cta2.start()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
import sys
import os
# 获取当前脚本的绝对路径
current_path = os.path.abspath(__file__)
root_path = os.path.abspath(os.path.join(current_path, "../../"))
# 将根目录添加到 sys.path
if root_path not in sys.path:
    sys.path.append(root_path)
from trade_mdforopenctp import PeopleQuantApi 
import time as tm
import zhuchannel
import asyncio
import traceback
import types
import polars
from datetime import datetime,time,date,timedelta
import copy
from typing import Dict, List, Optional, Tuple, Any

envs = {
    "7x24": {
        "td": "tcp://182.254.243.31:40001",
        "md": "tcp://182.254.243.31:40011",
        "user_id": "",
        "password": "",
        "broker_id": "9999",
        "authcode": "0000000000000000",
        "appid": "simnow_client_test",
        "user_product_info": "",
    },
    "电信1": {
        "td": "tcp://182.254.243.31:30001",
        "md": "tcp://182.254.243.31:30011",
        "user_id": "",
        "password": "",
        "broker_id": "9999",
        "authcode": "0000000000000000",
        "appid": "simnow_client_test",
        "user_product_info": "",
    },
    "电信2": {
        "td": "tcp://182.254.243.31:30002",
        "md": "tcp://182.254.243.31:30012",
        "user_id": "",
        "password": "",
        "broker_id": "9999",
        "authcode": "0000000000000000",
        "appid": "simnow_client_test",
        "user_product_info": "",
    },
    "移动": {
        "td": "tcp://182.254.243.31:30003",
        "md": "tcp://182.254.243.31:30013",
        "user_id": "",
        "password": "",
        "broker_id": "9999",
        "authcode": "0000000000000000",
        "appid": "simnow_client_test",
        "user_product_info": "",
    },
}
TradeFrontAddr="tcp://180.168.146.187:10101"   #交易前置地址
MdFrontAddr="tcp://101.230.209.178:53313"      #行情前置地址
TradeFrontAddr = envs["电信2"]["td"]
MdFrontAddr = envs["电信2"]["md"]
TradeFrontAddr = envs["7x24"]["td"]
MdFrontAddr = envs["7x24"]["md"]

#TradeFrontAddr = "tcp://121.37.80.177:20002" #openctp
#MdFrontAddr = "tcp://121.37.80.177:20004" #openctp

BROKERID="9999"   #期货公司ID
USERID=""   #账户
PASSWORD=""   #登录密码
APPID="simnow_client_test"   #客户端ID
AUTHCODE="0000000000000000"  #授权码

def cta(name,symbol_ctp,filename,logfile,**kw ):
    quotes, positions = {}, {}  #不同期权的行情和持仓
    symbole_infos = {}   #不同期权合约属性
    quote_future = pqapi.get_quote(symbol_ctp)        #获取合约行情
    UnderlyingInstrID = symbol_ctp #期权标的合约
    option_calls = pqapi.query_symbol_option(UnderlyingInstrID,"CALL") #全部看涨期权
    option_puts = pqapi.query_symbol_option(UnderlyingInstrID,"PUT")   #全部看跌期权
    option_call = option_calls[0] #最近到期日全部看涨期权
    option_put = option_puts[0]   #最近到期日全部看跌期权
    UpdateTime = quote_future.ctp_datetime #行情更新时间
    lot = 5 #下单手数
    balance = kw["balance"] if "balance" in kw else 0  #账户最低权益
    risk_ratio = kw["risk_ratio"] if "risk_ratio" in kw else 1  #账户风险度
    #报单次数、撤单次数、多空开仓成交总手数、自成交数、信息量       阈值
    orders_insert,orders_cancel,daylots,self_trade,order_exe = 50000,50000,100000,10000,100000
    while True:
        if UpdateTime != quote_future.ctp_datetime: #新行情推送
            UpdateTime = quote_future.ctp_datetime
            #权益足够,风险度足够,否则只平不开
            risk_control = account["Balance"] > balance and account["risk_ratio"] < risk_ratio
            orderrisk = pqapi.get_order_risk(symbol_ctp)
            order_enable = (orderrisk["order_count"] < orders_insert and orderrisk["cancel_count"] < orders_cancel and
                            orderrisk["open_volume"] < daylots and orderrisk["self_trade_count"] < self_trade and
                            orderrisk["order_exe"] < order_exe)
            #期货发出交易信号
            buy_up = 1  #多头信号
            sell_down = 1  #空头信号
            underlying_price = quote_future["LastPrice"] #基准价格
            if order_enable:
                if buy_up and risk_control:
                    #查询基准价格对应的虚值1档看涨期权
                    op_call = pqapi.get_option("CALL",underlying_price,price_level=-1,group_option=option_call)["option"]
                    if op_call not in positions: positions[op_call] = pqapi.get_position(op_call)   #获取合约持仓
                    if op_call not in quotes: quotes[op_call] = pqapi.get_quote(op_call)   #获取合约持仓
                    if op_call not in symbole_infos: symbole_infos[op_call] = pqapi.get_symbol_info(op_call)
                    position_call = positions[op_call]
                    quote_call = quotes[op_call]
                    if not position_call.pos_long:
                        price = quote_call["AskPrice1"]
                        r = pqapi.open_close(op_call,"kaiduo",lot,price,order_info='开仓')
                        print(r['kaiping'],r['shoushu'],r['junjia'],r["last_msg"],position_call )
                        if r['shoushu']:
                            new_df = polars.DataFrame([{'策略':name,'标的合约':symbol_ctp,'合约':op_call,'方向':r['kaiping'],'下单价格':r['price'],'下单数量':r['lot'],'成交价格':r['junjia'],'成交数量':r['shoushu'],'盘口挂单量':r['quote_volume'],'备注':r['order_info'],'利润点数':0,'利润金额':0, '日期':f'{UpdateTime}'}])
                            pqapi.trade_excel(new_df,'交易统计',filename,logfile) #保存全部成交记录
                if sell_down and risk_control:
                    #查询虚值1档看涨期权
                    op_put = pqapi.get_option("PUT",underlying_price,price_level=-1,group_option=option_put)["option"]
                    if op_put not in positions: positions[op_put] = pqapi.get_position(op_put)   #获取合约持仓
                    if op_put not in quotes: quotes[op_put] = pqapi.get_quote(op_put)   #获取合约持仓
                    if op_put not in symbole_infos: symbole_infos[op_put] = pqapi.get_symbol_info(op_put)
                    position_put = positions[op_put]
                    quote_put = quotes[op_put]
                    if not position_put.pos_long:
                        price = quote_put["AskPrice1"]
                        r = pqapi.open_close(op_put,"kaiduo",lot,price,order_info='开仓')
                        print(r['kaiping'],r['shoushu'],r['junjia'],r["last_msg"],position_put )
                        if r['shoushu']:
                            new_df = polars.DataFrame([{'策略':name,'标的合约':symbol_ctp,'合约':op_put,'方向':r['kaiping'],'下单价格':r['price'],'下单数量':r['lot'],'成交价格':r['junjia'],'成交数量':r['shoushu'],'盘口挂单量':r['quote_volume'],'备注':r['order_info'],'利润点数':0,'利润金额':0, '日期':f'{UpdateTime}'}])
                            pqapi.trade_excel(new_df,'交易统计',filename,logfile) #保存全部成交记录
                for s,p in positions.items():
                    if p.pos_long and abs(quotes[s].LastPrice - p.open_price_long) >= 1*symbole_infos[s]['PriceTick']:
                        price = quotes[s]["BidPrice1"]
                        open_price_long = p.open_price_long  #初始开仓价位
                        r = pqapi.open_close(s,"pingduo",p.pos_long,price,order_info='平仓')
                        print(r['kaiping'],r['shoushu'],r['junjia'],r["last_msg"],p )
                        if r['shoushu']:
                            profit_count = r['junjia'] - open_price_long #盈利价差
                            profit_money = profit_count * r['shoushu'] * symbole_infos[s]["VolumeMultiple"] #盈利金额
                            new_df = polars.DataFrame([{'策略':name,'标的合约':symbol_ctp,'合约':s,'方向':r['kaiping'],'下单价格':r['price'],'下单数量':r['lot'],'成交价格':r['junjia'],'成交数量':r['shoushu'],'盘口挂单量':r['quote_volume'],'备注':r['order_info'],'利润点数':profit_count,'利润金额':profit_money, '日期':f'{UpdateTime}'}])
                            pqapi.trade_excel(new_df,'交易统计',filename,logfile) #保存全部成交记录
                

#创建api实例
pqapi = PeopleQuantApi()
account = pqapi.get_account()            #获取账户资金

cta1 = zhuchannel.WorkThread(cta,args=('热卷测试','rb2601','程序1',r'C:\CTPLogs' ),kwargs={})
cta3 = zhuchannel.WorkThread(cta,args=('豆油测试','y2601', '程序1',r'C:\CTPLogs'),kwargs={})
cta1.start()
cta3.start()

local_timestamp = account.local_timestamp
while True:
    if local_timestamp != account.local_timestamp:
        local_timestamp = account.local_timestamp
        print(account,"\n")
    tm.sleep(10)
